#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      Mkgmov.py
#  brief:     Mkgmov.py
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import Mtools
import Mtools as mt

import os
import glob
from pNbody import *
from pNbody import iofunc
from pNbody.param import Params
import copy
import types
import numpy as np

import gzip


##########################################################################
#
#  Some info on the structure :
#
#  1) nb   : the full nbody object (main loop)
#
#  2) nbf  : normally set the display parameters, position of the observer
#            but could be used to set the component
#
#  3) nbfc : normally set the component to display and/or the value to dispplay
#
#      component['id'][0] == '@'
#        this is used to open simple nbody object (to draw a box, for example)
#      component['id'][0] == '#'
#        no selection
#      component['id']    == 'gas'
#        selection gas
#
#
#
#
#  a script may be called at level 2)
#
#    frame['ext_cmd'] = []
#    frame['ext_cmd'].append("""from plot_P import MakePlot""")
#    frame['ext_cmd'].append("""MakePlot([nbf],output)""")
#
#    it must save an output output, the latter is opened as an img
#    and appened to imgs
#
#
#  a script may be called at level 3)
#
#    1)
#    component['ext_cmd'] = []
#    component['ext_cmd'].append("""from plot_P import MakePlot""")
#    component['ext_cmd'].append("""MakePlot([nbfc],output)""")
#
#    it must save an output output, the latter is opened as an img
#    and appened to imgs
#
#
#    2) using
#    component['to_img']='script_name'
#
#
##########################################################################


#######################################################################
# some useful functions
#######################################################################


def ReadNbodyParameters(paramname):
    """
    read param from a parameter Nbody file
    """

    gparams = Params(paramname, None)

    param = {}
    # create new params
    for p in gparams.params:
        param[p[0]] = p[3]

    return param


def gzip_compress(file):

    f = open(file, 'rb')
    content = f.read()
    f.close()

    f = gzip.open(file + '.gz', 'wb')
    f.write(content)
    f.close()


#######################################################################
#
#	C L A S S   D E F I N I T I O N
#
#######################################################################


class Movie():

    def __init__(
            self,
            parameterfile='filmparam.py',
            format=None,
            imdir=None,
            timesteps=None,
            pio=False,
            compress=True,
            ifile=0):

        self.DEFAULT_TIMESTEPS = None
        self.DEFAULT_FORMAT = "fits"
        self.DEFAULT_IMDIR = "fits"
        self.DEFAULT_PIO = False
        self.DEFAULT_COMPRESS = True

        self.DEFAULT_SCALE = "log"
        self.DEFAULT_CD = 0.
        self.DEFAULT_MN = 0.
        self.DEFAULT_MX = 0.
        self.DEFAULT_PALETTE = "light"

        self.DEFAULT_SKIPPED_IO_BLOCKS = []

        self.parameterfile = parameterfile

        # read the parameter file
        self.read_parameterfile()

        # use options
        if format is not None:
            self.film['format'] = format
        if imdir is not None:
            self.film['imdir'] = imdir
        if timesteps is not None:
            self.film['timesteps'] = timesteps
        if pio is not None:
            self.film['pio'] = pio
        if compress is not None:
            self.film['compress'] = compress

        self.imdir = self.film['imdir']
        self.pio = self.film['pio']
        self.compress = self.film['compress']

        # deal with timesteps
        self.set_timesteps()

        self.ifile = ifile - 1
        self.file_offset = 0

        # whith gmkgmov, the init is only runned by the master, this line is
        # not needed...
        if mpi.mpi_IsMaster():

            if self.pio:
                self.pio = "yes"
            else:
                self.pio = "no"

            if self.parameterfile is None:
                print("you must specify a parameter file")
                sys.exit()

            if not os.path.exists(self.parameterfile):
                print(("file %s does not exists" % self.parameterfile))
                sys.exit()

            if not os.path.exists(self.imdir):
                os.mkdir(self.imdir)
            else:
                print(("directory %s exists !!!" % self.imdir))

                files = os.listdir(self.imdir)
                print(("the directory %s contains %d files" % (self.imdir,len(files))))

                # png files
                png_files = glob.glob(os.path.join(self.imdir, "*.png"))
                n_png_files = len(png_files)

                print(("the directory %s contains %d png files" % (self.imdir, n_png_files)))

                # fits files
                #fits_files = glob.glob(os.path.join(self.imdir, "*.fits"))
                fits_files = glob.glob(os.path.join(self.imdir, "*.fits.gz"))
                n_fits_files = len(fits_files)

                print(("the directory %s contains %d fits files" % (self.imdir, n_fits_files)))
                
                
                n_images_per_file = len(self.film['frames'][0]['components'])
                print()
                print(n_images_per_file)
                print()

                if n_png_files > 0:
                    self.file_offset = n_png_files/n_images_per_file  
                    
                if n_fits_files > 0:
                    self.file_offset = n_fits_files/n_images_per_file    
                    


    def info(self):
        print("INFO INFO INFO")
        # print self.film
        print((self.parameterfile))
        print((self.getftype()))
        print("INFO INFO INFO")

    def read_parameterfile(self):

        if not os.path.isfile(self.parameterfile):
            raise IOError(
                915, 'file %s not found ! Pease check the file name.' %
                (self.parameterfile))

        # import the parameter file as a module

        module_name = os.path.basename(os.path.splitext(self.parameterfile)[0])
        module_dir = os.path.dirname(self.parameterfile)

        if sys.path.count(module_dir) == 0:
            sys.path.append(module_dir)
        
        filmparam = __import__(module_name)

        self.film = filmparam.film

        # set some defaults
        if 'timesteps' not in self.film:
            self.film['timesteps'] = self.DEFAULT_TIMESTEPS
        if 'imdir' not in self.film:
            self.film['imdir'] = self.DEFAULT_IMDIR
        if 'format' not in self.film:
            self.film['format'] = self.DEFAULT_FORMAT
        if 'pio' not in self.film:
            self.film['pio'] = self.DEFAULT_PIO
        if 'compress' not in self.film:
            self.film['compress'] = self.DEFAULT_COMPRESS
        if 'skipped_io_blocks' not in self.film:
            self.film['skipped_io_blocks'] = self.DEFAULT_SKIPPED_IO_BLOCKS

        self.setftype(self.film['ftype'])

        # post process
        for i, frame in enumerate(self.film['frames']):
            frame['id'] = i

        # check
        # for frame in self.film['frames']:
        #  print frame['id']
        #  for component in frame['components']:
        #	print "  ",component['id']

    ##################################
    # time steps stuffs
    ##################################

    def set_timesteps(self):
        """
        define self.times (which is a list) based
        on the value contained in self.film['timesteps']
        """

        # self.times
        if self.film['timesteps'] == 'every':
            self.times = "every"

        elif isinstance(self.film['timesteps'], bytes):
            fname = self.film['timesteps']
            if not os.path.isfile(fname):
                raise IOError(
                    916,
                    'file %s not found ! Pease check the file name.' %
                    (fname))
            times = iofunc.read_ascii(fname, [0])[0]
            times = np.take(
                times,
                len(times) -
                1 -
                np.arange(
                    len(times)))  # invert order
            times = times.tolist()
            self.times = times

        elif isinstance(self.film['timesteps'], list):
            self.times = self.film['timesteps']

        elif isinstance(self.film['timesteps'], tuple):
            t0 = self.film['timesteps'][0]
            t1 = self.film['timesteps'][1]
            dt = self.film['timesteps'][2]
            times = np.arange(t0, t1, dt)
            times = np.take(
                times,
                len(times) -
                1 -
                np.arange(
                    len(times)))  # invert order
            times = times.tolist()
            self.times = times

        else:
            self.times = []

    def set_next_time(self):
        if self.times != "every":
            if len(self.times) > 0:
                self.times.pop()

    def get_next_time(self):
        if self.times == "every":
            return 0.0

        if len(self.times) == 0:
            return None
        else:
            return self.times[-1]

    def getftype(self):
        return self.ftype

    def setftype(self, ftype):
        self.ftype = ftype

    def get_skipped_io_blocks(self):
        return self.film["skipped_io_blocks"]

    def ApplyFilmParam(self, nb, film):
        
        # set time reference for this file
        exec("nb.tnow = %s" % film['time'])
        # exec1
        if 'exec' in film:
            if film['exec'] is not None:
                exec(film['exec'])

        # macro
        if 'macro' in film:
            if film['macro'] is not None:
                exec(
                    compile(
                        open(
                            film['macro']).read(),
                        film['macro'],
                        'exec'))

        return nb

    def ApplyFrameParam(self, nb, frame):

        global nbf
        
        nbf = nb

        # exec
        if 'exec' in frame:
            if frame['exec'] is not None:
                exec(frame['exec'])

        # macro
        if 'macro' in frame:
            if frame['macro'] is not None:
                exec(
                    compile(
                        open(
                            frame['macro']).read(),
                        frame['macro'],
                        'exec'))

        return nbf

    def ApplyComponentParam(self, nbf, component):

        global nbfc
        
        if component['id'][0] == '@':
            # here, all tasks must have an object containing all particles
            # ok, but not in the right order !!!
            nbfc = Nbody(componentid, self.getftype())
            nbfc = sorted(nbfc.SendAllToAll())
            nbfc.componentid = component['id']  # [1:]
        elif component['id'][0] == '#':
            nbfc = nbf
            nbfc.componentid = component['id']  # [1:]
        else:
            nbfc = nbf.select(component['id'])
            nbfc.componentid = component['id']

        # exec
        if 'exec' in component:
            if component['exec'] is not None:
                exec(component['exec'])

        # macro
        if 'macro' in component:
            if component['macro'] is not None:
                exec(
                    compile(
                        open(
                            component['macro']).read(),
                        component['macro'],
                        'exec'))

        # print "------------------------"
        # print min(nbfc.u),max(nbfc.u)
        # print min(nbfc.rho),max(nbfc.rho)
        # print min(nbfc.tpe),max(nbfc.tpe)
        # print "temperature",min(nbfc.T()),max(nbfc.T())
        # print nbfc.nbody
        # print min(nbfc.rsp),max(nbfc.rsp)
        # print "------------------------"

        return nbfc

    def dump(self, dict):

        # exctract dict
        atime = dict['atime']
        pos = dict['pos']

        # create nbody object
        nb = Nbody(pos=pos, ftype='gadget')
        nb.atime = atime

        # add other arrays
        if 'vel' in dict:
            nb.vel = dict['vel']

        if 'num' in dict:
            nb.num = dict['num']

        if 'mass' in dict:
            nb.mass = dict['mass']

        if 'tpe' in dict:
            nb.tpe = dict['tpe']

        if 'u' in dict:
            nb.u = dict['u']
        if 'rho' in dict:
            nb.rho = dict['rho']
        if 'rsp' in dict:
            nb.rsp = dict['rsp']

        if 'metals' in dict:
            nb.metals = dict['metals']

            #!!!
            nb.flag_chimie_extraheader = 1
            nb.ChimieNelements = 5
            nb.flag_metals = 5
            nb.ChimieSolarMassAbundances = {}
            nb.ChimieSolarMassAbundances['Fe'] = 0.00176604

        nb.init()

        # print "################################"
        # print "writing qq.dat"
        # print "################################"
        # nb.rename('qq.dat')
        # nb.write()

        self.dumpimage(nb=nb)

    def dumpimage(self, nb=None, file=None):

        # increment counter
        self.ifile += 1

        # skip file if needed
        if self.ifile < self.file_offset:
            if mpi.mpi_IsMaster():
                print(("skipping file %04d" % self.ifile))

            return

        # for each frame one can create an img
        imgs = []

        if file is not None:
            # print "(task=%04d) reading
            # "%(mpi.ThisTask),file,self.getftype(),self.pio

            if mpi.mpi_IsMaster():
                print()
                print("#############################################################")
                print(("reading...", file))
                print("#############################################################")

            nb = Nbody(file, ftype=self.getftype(), pio=self.pio,
                       skipped_io_blocks=self.get_skipped_io_blocks())
        else:
            if nb is None:
                raise Exception(
                    "you must specify at least a file or give an nbody object")

        film = self.film

        nb = self.ApplyFilmParam(nb, film)

        for frame in film['frames']:
            nbf = self.ApplyFrameParam(nb, frame)

            if 'ext_cmd' in frame:
                if len(frame['ext_cmd']) > 0:
                    #################################################
                    # 1) use an outer script to create an img		(this is a bit redundant with 2.2, see below )
                    #################################################
                    output = "/tmp/%015d.png" % (int(np.random.random() * 1e17))

                    for cmd in frame['ext_cmd']:
                        exec(cmd)

                    if mpi.mpi_IsMaster():
                        img = Image.open(output)
                        imgs.append(img)

                        if os.path.exists(output):
                            os.remove(output)

                continue

            # composition parameters

            if 'cargs' in frame:
                if len(frame['cargs']) != 0:
                    frame['compose'] = True
                    datas = []
                else:
                    frame['compose'] = False
            else:
                frame['cargs'] = []
                frame['compose'] = False

            for component in frame['components']:
                if mpi.mpi_IsMaster():
                    print("------------------------")
                    print(("component", component['id']))
                    print("------------------------")

                nbfc = self.ApplyComponentParam(nbf, component)

                # find the observer position
                # 1) from params
                # 2) from pfile
                # 3) from tdir
                # and transform into parameter

                if frame['tdir'] is not None:

                    tfiles = sorted(
                        glob.glob(
                            os.path.join(
                                frame['tdir'],
                                "*")))

                    bname = os.path.basename(file)

                    tfiles_for_this_file = []
                    for j in range(len(tfiles)):
                        tfile = "%s.%05d" % (bname, j)
                        # tfile = "%s.0.%05d"%(bname,j) # old or new version ?
                        tmp_tfile = os.path.join(frame['tdir'], tfile)

                        if os.path.exists(tmp_tfile):
                            tfiles_for_this_file.append(tmp_tfile)

                elif frame['pfile'] is not None:

                    if not os.path.isfile(frame['pfile']):
                        print((
                            "parameter file %s does not exists(1)..." %
                            (frame['pfile'])))

                    # read from pfile defined in frame
                    param = ReadNbodyParameters(frame['pfile'])
                    tfiles_for_this_file = [None]

                else:

                    # take frame as parameter
                    param = copy.copy(frame)
                    tfiles_for_this_file = [None]

                # loop over different oberver positions for this file
                for iobs, tfile in enumerate(tfiles_for_this_file):

                    if tfile is not None:
                        if mpi.mpi_IsMaster():
                            print(("  using tfile : %s" % (tfile)))
                        param = ReadNbodyParameters(tfile)

                    # add parameters defined by user in the parameter file
                    for key in list(component.keys()):
                        param[key] = component[key]

                    # set image shape using frame
                    param['shape'] = (frame['width'], frame['height'])

                    # compute map
                    mat = nbfc.CombiMap(param)

                    if mpi.mpi_IsMaster():

                        if frame['compose']:
                            datas.append(mat)

                        if 'ext_cmd' in component:
                            #################################################
                            # 1) use an outer script to create an img
                            #################################################
                            if len(component['ext_cmd']) > 0:

                                output = "/tmp/%015d.png" % (
                                    int(np.random.random() * 1e17))

                                for cmd in component['ext_cmd']:
                                    exec(cmd)

                                if mpi.mpi_IsMaster():
                                    img = Image.open(output)
                                    imgs.append(img)

                                    if os.path.exists(output):
                                        os.remove(output)

                        elif self.film["format"] == "fits":
                            #################################
                            # 1) save fits file
                            #################################
                            output = '%04d_%04d-%s-%06d.fits' % (
                                self.ifile, frame['id'], component['id'], iobs)
                            output = os.path.join(self.imdir, output)
                            print((nb.atime, output))

                            if os.path.exists(output):
                                os.remove(output)

                            header = [('TIME', nb.tnow, 'snapshot time')]
                            iofunc.WriteFits(
                                np.transpose(mat), output, extraHeader=header)

                            # compress
                            if self.compress:
                                gzip_compress(output)
                                os.remove(output)

                        elif self.film["format"] == "png":
                            #################################
                            # 2) output png file or ...
                            #################################

                            output = '%04d_%04d-%s-%06d.png' % (
                                self.ifile, frame['id'], nbfc.componentid, iobs)

                            output = os.path.join(self.imdir, output)
                            print((nb.atime, output))

                            # here, we should use component['scale'] ... not
                            # frame['scale'], no ?

                            if 'scale' not in frame:
                                frame['scale'] = self.DEFAULT_SCALE
                            if 'cd' not in frame:
                                frame['cd'] = self.DEFAULT_CD
                            if 'mn' not in frame:
                                frame['mn'] = self.DEFAULT_MN
                            if 'mx' not in frame:
                                frame['mx'] = self.DEFAULT_MX

                            if 'palette' not in frame:
                                frame['palette'] = self.DEFAULT_PALETTE

                            matint, mn_opt, mx_opt, cd_opt = set_ranges(
                                mat, scale=frame['scale'], cd=frame['cd'], mn=frame['mn'], mx=frame['mx'])
                            frame['mn'] = mn_opt
                            frame['mx'] = mx_opt
                            frame['cd'] = cd_opt
                            img = get_image(
                                matint, palette_name=frame['palette'])
                            img.save(output)

                            print((frame['mn'], frame['mx'], frame['cd']))

                        # need to create an img
                        if 'to_img' in component:

                            if component['to_img']:
                                ##########################################
                                # 2.1) create an img and apply commmands
                                ##########################################

                                # get params
                                if 'scale' not in component:
                                    component['scale'] = self.DEFAULT_SCALE
                                if 'cd' not in component:
                                    component['cd'] = self.DEFAULT_CD
                                if 'mn' not in component:
                                    component['mn'] = self.DEFAULT_MN
                                if 'mx' not in component:
                                    component['mx'] = self.DEFAULT_MX
                                if 'palette' not in component:
                                    component['palette'] = self.DEFAULT_PALETTE

                                matint, mn_opt, mx_opt, cd_opt = set_ranges(
                                    mat, scale=component['scale'], cd=component['cd'], mn=component['mn'], mx=component['mx'])
                                img = get_image(
                                    matint, palette_name=component['palette'])

                                print((mn_opt, mx_opt, cd_opt))

                                # here we can add img commands....
                                if 'img_cmd' in component:
                                    if len(component['img_cmd']) > 0:

                                        for cmd in component['img_cmd']:
                                            exec(cmd)

                                # append img to list
                                img.atime = nb.atime
                                imgs.append(img)

                            elif isinstance(component['to_img'], bytes):
                                ##########################################
                                # 2.2) use an outer script to create an img from mat
                                ##########################################

                                output = "/tmp/%015d.png" % (
                                    int(np.random.random() * 1e17))

                                # get params
                                if 'scale' not in component:
                                    component['scale'] = self.DEFAULT_SCALE
                                if 'cd' not in component:
                                    component['cd'] = self.DEFAULT_CD
                                if 'mn' not in component:
                                    component['mn'] = self.DEFAULT_MN
                                if 'mx' not in component:
                                    component['mx'] = self.DEFAULT_MX
                                if 'palette' not in component:
                                    component['palette'] = self.DEFAULT_PALETTE

                                component['atime'] = nbfc.atime

                                mk = __import__(component['to_img'])
                                mk.MkImage(mat, output, component)

                                img = Image.open(output)
                                imgs.append(img)

                                os.remove(output)

            del nbf

            #######################
            # compose components
            #######################

            if frame['compose']:

                if mpi.mpi_IsMaster():

                    img, cargs = Mtools.fits_compose_colors_img(
                        datas, frame['cargs'])

                    # save
                    #output = '%04d_%04d.png'%(self.ifile,frame['id'])
                    #output = os.path.join(self.imdir,output)
                    # img.save(output)

                    # append img to list
                    img.atime = nb.atime
                    imgs.append(img)

        del nb

        #######################
        # compose frames
        #######################

        if mpi.mpi_IsMaster():
            if 'img_cmd' in film:
                if len(film['img_cmd']) > 0:

                    for cmd in film['img_cmd']:
                        exec(cmd)

                    output = '%04d.png' % (self.ifile)
                    output = os.path.join(self.imdir, output)
                    img.save(output)

                    img.i = 0



###########################################################################################
#  package:   pNbody
#  file:      bnbf.py
#  brief:
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

##########################################################################
#
# DEFAULT CLASS
#
##########################################################################

from pNbody import error


class Nbody_default:
    """
    This class is usefull to create an empty Nbody object
    """

    def _init_spec(self):
        """
        Initialize specific variable for this format
        """
        pass

    def get_excluded_extension(self):
        """
        Return a list of file to avoid when extending the default class.
        """
        return "all"

    def check_spec_ftype(self):
        """
        Check the format type.
        As this format is only virtual, allways return an error.
        """
        raise error.FormatError("default")

    def get_read_fcts(self):
        """
        returns the functions needed to read a file.
        """
        return [self.read_particles]

    def get_write_fcts(self):
        """
        returns the functions needed to write a file.
        """
        return [self.write_particles]

    def get_default_spec_vars(self):
        """
        Specific variables to this format
        """
        return {}

    def get_default_spec_array(self):
        """
        return specific array default values for the class
        """
        return {}

    def read_particles(self, f):
        """
        Function that read particles.
        """
        pass

    def write_particles(self, f):
        """
        Function that write particles.
        """
        pass



###########################################################################################
#  package:   pNbody
#  file:      default_extension.py
#  brief:     Default Nbody class extension
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import types
import os
import numpy as np

from copy import copy, deepcopy

import pNbody
from pNbody import thermodyn, mapping, myNumeric
from pNbody import ctes, libutil, FLOAT, nbodymodule
from pNbody import coolinglib, geometry, treelib, libgrid, libdisk
from pNbody import fourier
import pNbody.iofunc as pnio
import pNbody.mpiwrapper as mpi

from scipy.interpolate import interp1d

class _NbodyDefault:



    #################################
    #
    # Thermodynamic functions
    #
    #################################

    def U(self):
        """
        Return the gas specific energy of the model.
        The output is an nx1 float array.
        """
        return self.u

    def MeanWeight(self):
        """
        Return the mean weight of a model, taking into account
        heating by UV source.
        The output is an nx1 float array.
        """

        xi = self.unitsparameters.get('xi')
        Redshift = 1. / self.atime - 1.

        UnitDensity_in_cgs = self.localsystem_of_units.get_UnitDensity_in_cgs()
        UnitEnergy_in_cgs = self.localsystem_of_units.get_UnitEnergy_in_cgs()
        UnitMass_in_g = self.localsystem_of_units.get_UnitMass_in_g()
        HubbleParam = self.hubbleparam

        # 0) convert into cgs
        Density = self.rho.astype(
            float) * UnitDensity_in_cgs * (HubbleParam * HubbleParam) / self.atime**3
        Egyspec = self.u.astype(float) * UnitEnergy_in_cgs / UnitMass_in_g

        # 1) compute mu
        MeanWeight, Lambda = coolinglib.cooling(Egyspec, Density, xi, Redshift)

        return MeanWeight.astype(np.float32)

    def A(self):
        """
        Return the gas entropy of the model.
        The output is an nx1 float array.
        """
        gamma = self.unitsparameters.get('gamma')
        xi = self.unitsparameters.get('xi')
        ionisation = self.unitsparameters.get('ionisation')
        mu = thermodyn.MeanWeight(xi, ionisation)
        mh = ctes.PROTONMASS.into(self.localsystem_of_units)
        k = ctes.BOLTZMANN.into(self.localsystem_of_units)
        thermopars = {"k": k, "mh": mh, "mu": mu, "gamma": gamma}

        A = np.where(
            (self.u > 0),
            thermodyn.Aru(
                self.Rho(),
                self.u,
                thermopars),
            0)
        return A

    def P(self):
        """
        Return the gas pressure of the model.
        The output is an nx1 float array.
        """
        gamma = self.unitsparameters.get('gamma')
        xi = self.unitsparameters.get('xi')
        ionisation = self.unitsparameters.get('ionisation')
        mu = thermodyn.MeanWeight(xi, ionisation)
        mh = ctes.PROTONMASS.into(self.localsystem_of_units)
        k = ctes.BOLTZMANN.into(self.localsystem_of_units)
        thermopars = {"k": k, "mh": mh, "mu": mu, "gamma": gamma}

        P = np.where(
            (self.u > 0),
            thermodyn.Pru(
                self.Rho(),
                self.u,
                thermopars),
            0)
        return P

    def Ne(self):
        """
        Return the electron density of the model.
        The output is an nx1 float array.
        """

        xi = self.unitsparameters.get('xi')
        ionisation = self.unitsparameters.get('ionisation')
        mh = ctes.PROTONMASS.into(self.localsystem_of_units)

        thermopars = {"mh": mh, "Xi": xi, "ionisation": ionisation}
        ne = thermodyn.ElectronDensity(self.Rho(), thermopars)

        return ne

    def S(self):
        """
        Return the `entropy` of the model, defined as
        S = T * Ne^(1-gamma)
        The output is an nx1 float array.
        """

        gamma = self.unitsparameters.get('gamma')

        s = self.T() * self.Ne()**(1. - gamma)

        return s

    def Lum(self):
        """
        Return the luminosty of the model, defined as
        Lum = m*u/Tcool = m*Lambda/rho

        The output is an nx1 float array.
        """

        Lum = self.mass * self.u / self.Tcool()

        return Lum



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      isothermal.py
#  brief:     Defines isothermal profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


"""
isothermal model
"""

import numpy as np


def Potential(rho0, a, r, G=1.):
    """
    Potential
    """
    return 4*np.pi * G * rho0 * a**2 * np.log(r/a)


def Density(rho0, a, r, G=1.):
    """
    Density
    """
    return rho0 * a**2/r**2
    


def Vcirc(rho0, a, r, G=1.):
    """
    circular velocity
    """
        
    def vc():
      return 4*np.pi*G*rho0*a**2
    
    
    if type(r) == np.ndarray:
      return np.ones(len(r)) * vc()
    
    else:
      return vc()
    
    



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:     integration_utils.py
#  brief:     Contains utilities the orbit_integration scripts
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Darwin Roduit <darwin.roduit@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################
from pNbody.mass_models import MWPotential2014 as MW



def MW_time_evolution_factor(time, slope=0.09):
    """Time dependent multiplicative factor to change the mass off the
    MWPotential as time evolves. We assume a linear growth in time. This is
    roughly valid during the last 10 Gyr. """
    return slope*time + 1


def GradPot_time_evolved(rho_0, r_s, M_disk, a, b, alpha,
                             r_c, amplitude, r_1, f_1, f_2, f_3, position, time, G=1):
    """Gradient of MW potential (time evolving) in time in carthesian
    coordinates. For orbit integratrion."""

    # We add a time evolved component
    factor = MW_time_evolution_factor(time)

    # Total gradient
    grad_Phi = factor*MW.GradPot(rho_0, r_s, M_disk, a, b, alpha, r_c,
                                  amplitude, r_1, f_1, f_2, f_3, position, time, G)

    return grad_Phi



#!/usr/bin/env python3

###########################################################################################
#  package:   pNbody
#  file:      libSSPluminosity.py
#  brief:     SSP luminosities
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
###########################################################################################

import scipy
import scipy.interpolate
import numpy as np

from pNbody import *


class SSPLuminosities:

    def __init__(self, file,band,luminosity=False):

        self.file = file
        self.band = band
        self.luminosity = luminosity

        # read file and crate self.data
        self.Read()

        # create the matrix
        self.CreateMatrix()

    def Read(self):
        """
        read file and create a data table
        """
        pass

    def CreateMatrix(self):
        """
        from data extract
        metalicites (zs)
        ages (ages)
        and ML (vs)
        """

        pass

    def CreateInterpolator(self):
        """
        from the matrix self.MatLv, create a spline interpolator
        """
        self.spl = scipy.interpolate.RectBivariateSpline(
            self.Zs, self.Ages, self.MatLv)

    def ExtrapolateMatrix(self, order=1, zmin=-5, zmax=2, nz=50, s=0, extrapolate=True):
        """
        extrapolate the matrix self.MatLv in 1d (using spline), along the Z axis
        The function create a new self.MatLv and self.Zs
        """

        # if extrapolated is true, we extrapolate values further out
        # elsewhere we use the neareast value (keep the value constant)
        if extrapolate:
          ext=0 
        else:
          ext=3   

        xx = np.linspace(zmin, zmax, nz)

        newMatLv = np.zeros((len(xx), len(self.Ages)))

        for i in np.arange(len(self.Ages)):

            Ls = self.MatLv[:, i]

            # 1d spline interpolation
            x = self.Zs
            y = Ls

            tck = scipy.interpolate.fitpack.splrep(x, y, k=order, s=s)
            yy = scipy.interpolate.fitpack.splev(xx, tck, ext=ext)

            newMatLv[:, i] = yy

        self.Zs = xx
        self.MatLv = newMatLv

    def Extrapolate2DMatrix(
            self,
            zmin=-10,
            zmax=2,
            nz=256,
            agemin=None,
            agemax=None,
            nage=256):
        if agemin is None:
            agemin = min(self.Ages)
        if agemax is None:
            agemax = max(self.Ages)

        self.Zs = np.linspace(zmin, zmax, nz)
        self.Ages = 10**np.linspace(np.log10(agemin), np.log10(agemax), nage)
        #self.Ages  = np.linspace((agemin),(agemax),nage)

        self.MatLv = self.Luminosity(self.Zs, self.Ages)

    def GetAgeIndexes(self, Ages):
        """
        Get the indexes of the nearest values of self.Ages from Ages
        """
        return self.Ages.searchsorted(Ages)

    def GetZIndexes(self, Zs):
        """
        Get the indexes of the nearest values of self.Zs from Zs
        """
        return self.Zs.searchsorted(Zs)

    def Luminosity(self, Zs, Ages):
        """
        return an interpolated value of Luminosity using self.slp
        from a given Zs and Ages
        """

        MatLvi = self.spl(Zs, Ages)
        return MatLvi

    def Luminosities(self, Zs, Ages):
        """
        return an interpolated value of Luminosity using self.slp
        from a given Zs and Ages
        """

        i = self.Zs.searchsorted(Zs)
        j = self.Ages.searchsorted(Ages)
        i = i.clip(0, len(self.Zs) - 1)
        j = j.clip(0, len(self.Ages) - 1)

        return self.MatLv[i, j]



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      shell.py
#  brief:     Defines shell profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


"""
shell model
"""

import numpy as np


def Potential(M, R, r, G=1.):
    """
    Shell Potential
    """
    
    P1 = -G * M / R     # r < R
    P2 = -G * M / r     # r > R
    
    
    return np.where(r<R,P1,P2)
    
    
    
def dPotential(M, R, r, G=1.):
    """
    Shell first derivative of Potential
    """
    
    dP1 = 0              # r < R
    dP2 = G * M / r**2   # r > R
    
    return np.where(r<R,dP1,dP2)
    
    
    
    
        
    
    
    
    
    



###########################################################################################
#  package:   Mockimgs
#  file:      los.py
#  brief:     line of sight class
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
from .. import iofunc as io
from .utils import align
from .parameters import Parameters

def get_axes(n=7,random=False,irand=0):
  
  if random:
    np.random.seed(irand)
    random2 = np.random.random(n)
    random3 = np.random.random(n)
  else:

    # this is another way to define n1 and n2
    # n2 = int(1 + np.sqrt(1-0.5*(3-n)))
    # n1 = 2*(n2-1)
    # from the requirement that n = 2*n2**2 -4*n2 + 3 
    # which comes from  n = n1*n2 - (n1-1)
        
    n = int(np.sqrt(n/2))
    #n1 = 2*n   
    n1 = n      
    n2 = n+1
    #random2 = np.linspace(0,1,n1+1)[:-1]   
    random2 = np.linspace(0,1,n1+1)[:-1]/2  # limit to pi
    
    random3 = np.linspace(0,0.5,n2)    
    random2,random3 = np.meshgrid(random2,random3)
    random2 = random2.ravel()
    random3 = random3.ravel()
    
    # keep only one 90 deg in theta
    random2 = random2[(n1-1):]
    random3 = random3[(n1-1):]
    

  phi = random2 * np.pi * 2.
  costh = 1. - 2. * random3

  sinth = np.sqrt(1. - costh**2)

  x = sinth * np.cos(phi)
  y = sinth * np.sin(phi)
  z = costh
  
  pos = np.transpose(np.array([x, y, z]))
  
  return pos


def get_axes_from_grid(n_phi,n_theta,d_phi,d_theta,los):
  """
  get a list of axes centred on los and 
  equally distributed in azimuth and elevation 
  
  n_phi   :   number of divisions in azimuthal angle
  n_theta :   number of divisions in elevation angle
  d_phi   :   azimuthal variation in degree [-d_phi,-d_phi]
  d_theta :   elevation variation in degree [-d_theta,-d_theta]
  
  """
  
  thetas = np.pi/180 * np.linspace(-d_theta,d_theta,n_theta)
  phis   = np.pi/180 * np.linspace(-d_phi  ,d_phi  ,n_phi)
  
  # the ref axis is 1,0,0
  axes = []
  for theta in thetas:
    for phi in phis:
      x = np.cos(theta)*np.cos(phi)
      y = np.cos(theta)*np.sin(phi)
      z = np.sin(theta)
      axes.append([x,y,z])
  
  # convert to numpy array
  pos = np.array(axes)
  # rotate to centre the final direction
  pos = align(pos,axis=los,axis_ref=[1,0,0])
  # back to a list
  axes = pos.tolist()
  
  return axes





  

def SetLineOfSights(n=1,random=False):

  if n==1:
    los_axes=[[0,0,1]]
  else:    
    los_axes = get_axes(n,random)  
    
  return los_axes


class LineOfSights():
  """
  A line of sights object
  
  not that a line of sight of [1,0,0] means that the observer
  is in [1,0,0] and look towards [-1,0,0] 
  
  parameters
  ----------
  
  n         : the number of line of sight
  random    : if the line of sight are random or not
  irand     : the random seed
  los       : explicitly provide a line of sight
  
  
  """
  def __init__(self,opt=None,params=None,parameter_file=None):
    
    # init default parameters
    self._init_default_parameters()
    
    # 1: set parameters from a yml file
    if parameter_file is not None:
      pdic = io.ReadYaml(parameter_file,'LineOfSights')
      self._p.update_from_dic(pdic) 

    # 2: set parameters from the dictionary params
    if params is not None:
      self._p.update_from_dic(params) 

    # 3: set parameters from options
    if opt is not None:
      self._p.update_from_options(opt,skip_None=True) 
    
    
    
    # init idx
    self.idx = -1
        
    # treat different cases
    
    # create from a grid
    if self._p.grid is not None:
      n_phi   = self._p.grid["n_phi"]
      n_theta = self._p.grid["n_theta"]
      d_phi   = self._p.grid["d_phi"]
      d_theta = self._p.grid["d_theta"] 
      los     = self._p.los     
      self.axes = get_axes_from_grid(n_phi,n_theta,d_phi,d_theta,los)
      return  
  
    # if the line of sight is given
    if self._p.los is not None:
      # use the los which is provided
      self.axes=[self._p.los]
      return
    
    # only one line of sight is needed
    if self._p.nlos==None or self._p.nlos==1:
      self.axes=[[0,0,1]]
      return
      
    # instead, get either random axes or axes on an hemisphere  
    self.axes = get_axes(self._p.nlos,random=self._p.random_los,irand=self._p.irand)  
    
    
    
  def _init_default_parameters(self):
    '''
    define parameters as well as their default value
    '''
    self._p = Parameters()
    self._p.nlos        = None      # number of line of sights
    self._p.random_los  = False     # random line of sights
    self._p.irand       = 0         # random seed
    self._p.los         = None      # line of sight
    self._p.grid        = None      # grid parameters to generate a los
  
  def print_parameters(self):
    print(self._p)  

    
  def __iter__(self):
    return self

  def __next__(self):
    if self.idx == len(self.axes)-1:
      raise StopIteration
    self.idx = self.idx + 1
    return self.axes[self.idx]

  def info(self):
    for ax in self.axes:
      print(ax)
  
  def n(self):
    """
    return the number of line of sights
    """
    return len(self.axes)  
  
  def list(self):
    """
    return the list of line of sight
    """
    return self.axes
        
  def get(self):
    """
    return the next line of sight
    """
    return self.axes[self.idx]
    
    


  



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      kernels.py
#  brief:     kernels definitions
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

"""
kernel implementation
see : Dehnen & Aly 2012
"""


import numpy as np

# normalization constants

C_cub_spline_1D  = 8/3.
C_qua_spline_1D  = 5**5/768
C_qui_spline_1D  = 3**5/40

C_cub_spline_2D  = 80./(7*np.pi)
C_qua_spline_2D  = 3*5**6/(2398*np.pi)
C_qui_spline_2D  = 7*3**7/(478*np.pi)

C_cub_spline_3D  = 16/np.pi
C_qua_spline_3D  = 5**6/(512*np.pi)
C_qui_spline_3D  = 3**7/(40*np.pi)

# H/h ratio
# h is the smoothing scale
# H is the kernel-support radius

Hh_cub_spline_1D = 1.732051
Hh_qua_spline_1D = 1.936492
Hh_qui_spline_1D = 2.121321

Hh_cub_spline_2D = 1.778002
Hh_qua_spline_2D = 1.977173
Hh_qui_spline_2D = 2.158131

Hh_cub_spline_3D = 1.825742
Hh_qua_spline_3D = 2.018932
Hh_qui_spline_3D = 2.195775



def cub_spline(r):
  '''
  Cubic spline 
  not normalized
  '''
  
  w1 = (1-r)
  w1 = np.where(w1>0,w1,0)**3
  
  w2 = (0.5-r) 
  w2 = -4*np.where(w2>0,w2,0)**3 
  
  w = w1+w2

  return w
  

def qua_spline(r):
  '''
  Quartic spline
  not normalized 
  '''  
    
  w1 = (1-r)
  w1 = np.where(w1>0,w1,0)**4

  w2 = (3/5-r) 
  w2 = -5*np.where(w2>0,w2,0)**4  

  w3 = (1/5-r) 
  w3 = +10*np.where(w3>0,w3,0)**4  
  
  w = w1+w2+w3

  return w


def qui_spline(r):
  '''
  Quintic spline
  not normalized 
  '''  
    
  w1 = (1-r)
  w1 = np.where(w1>0,w1,0)**5

  w2 = (2/3-r) 
  w2 = -6*np.where(w2>0,w2,0)**5  

  w3 = (1/3-r) 
  w3 = +15*np.where(w3>0,w3,0)**5  
    
  w = w1+w2+w3

  return w



def Wcub_1D(r,h):
  """
  Normalized 1D cubic kernel 
  
  h : the smoothing scale (2 sigma)
  """
  h = h*Hh_cub_spline_1D  # h becomes now H, the kernel-support radius
  return   C_cub_spline_1D * cub_spline(r/h)/h
  
def Wqua_1D(r,h):
  """
  Normalized 1D quadratic kernel 
  
  h : the smoothing scale (2 sigma)
  """  
  h = h*Hh_qua_spline_1D  # h becomes now H, the kernel-support radius
  return   C_qua_spline_1D * qua_spline(r/h)/h  
  
def Wqui_1D(r,h):
  """
  Normalized 1D quintic kernel 
  
  h : the smoothing scale (2 sigma)
  """    
  h = h*Hh_qui_spline_1D  # h becomes now H, the kernel-support radius
  return   C_qui_spline_1D * qui_spline(r/h)/h  

def Wcub_2D(r,h):
  """
  Normalized 2D cubic kernel 
  
  h : the smoothing scale (2 sigma)
  """  
  h = h*Hh_cub_spline_2D  # h becomes now H, the kernel-support radius
  return   C_cub_spline_2D * cub_spline(r/h)/h**2
  
def Wqua_2D(r,h):
  """
  Normalized 2D quadratic kernel 
  
  h : the smoothing scale (2 sigma)
  """    
  h = h*Hh_qua_spline_2D  # h becomes now H, the kernel-support radius
  return   C_qua_spline_2D * qua_spline(r/h)/h**2  
  
def Wqui_2D(r,h):
  """
  Normalized 2D quintic kernel 
  
  h : the smoothing scale (2 sigma)
  """    
  h = h*Hh_qui_spline_2D  # h becomes now H, the kernel-support radius
  return   C_qui_spline_2D * qui_spline(r/h)/h**2    
  
def Wcub_3D(r,h):
  """
  Normalized 3D cubic kernel 
  
  h : the smoothing scale (2 sigma)
  """
  h = h*Hh_cub_spline_3D  # h becomes now H, the kernel-support radius  
  return   C_cub_spline_3D * cub_spline(r/h)/h**3
  
def Wqua_3D(r,h):
  """
  Normalized 3D quadratic kernel 
  
  h : the smoothing scale (2 sigma)
  """    
  h = h*Hh_qua_spline_3D  # h becomes now H, the kernel-support radius
  return   C_qua_spline_3D * qua_spline(r/h)/h**3  
  
def Wqui_3D(r,h):
  """
  Normalized 3D quintic kernel 
  
  h : the smoothing scale (2 sigma)
  """  
  h = h*Hh_qui_spline_3D  # h becomes now H, the kernel-support radius
  return   C_qui_spline_3D * qui_spline(r/h)/h**3   
  


######################################
# serial C versions
######################################

def Wcub_1Ds(r,h):
  '''
  Normalized 1D cubic kernel : serial version
  
  h : the smoothing scale (2 sigma)
  '''
  
  # h becomes now H, the kernel-support radius
  h = h*Hh_cub_spline_1D  
  
  # scale the radius
  r = r/h
  
  # normalized kernel part
  
  w = 0
  w = w +  max(1-r,0)**3
  w = w -4*max(0.5-r,0)**3  
  
  w = w *C_cub_spline_1D/h
  
  return w


def Wqua_1Ds(r,h):
  '''
  Normalized 1D quadratic kernel : serial version 
  
  h : the smoothing scale (2 sigma)
  '''
  
  # h becomes now H, the kernel-support radius
  h = h*Hh_qua_spline_1D  
  
  # scale the radius
  r = r/h
  
  # normalized kernel part
  
  w = 0
  w = w +   max(1-r,0)**4
  w = w -5 *max(3/5-r,0)**4  
  w = w +10*max(1/5-r,0)**4
  
  w = w *C_qua_spline_1D/h
  
  return w  

def Wqui_1Ds(r,h):
  '''
  Normalized 1D quintic kernel : serial version 
  
  h : the smoothing scale (2 sigma)
  '''
  
  # h becomes now H, the kernel-support radius
  h = h*Hh_qui_spline_1D  
  
  # scale the radius
  r = r/h
  
  # normalized kernel part
  
  w = 0
  w = w +   max(1-r,0)**5
  w = w -6 *max(2/3-r,0)**5  
  w = w +15*max(1/3-r,0)**5
  
  w = w *C_qui_spline_1D/h
  
  return w    
  


def Wcub_2Ds(r,h):
  '''
  Normalized 2D cubic kernel : serial version 
  
  h : the smoothing scale (2 sigma)
  '''
  
  # h becomes now H, the kernel-support radius
  h = h*Hh_cub_spline_2D  
  
  # scale the radius
  r = r/h
  
  # normalized kernel part
  
  w = 0
  w = w +  max(1-r,0)**3
  w = w -4*max(0.5-r,0)**3  
  
  w = w *C_cub_spline_2D/(h*h)
  
  return w


def Wqua_2Ds(r,h):
  '''
  Normalized 2D quadratic kernel : serial version 
  
  h : the smoothing scale (2 sigma)
  '''
  
  # h becomes now H, the kernel-support radius
  h = h*Hh_qua_spline_2D  
  
  # scale the radius
  r = r/h
  
  # normalized kernel part
  
  w = 0
  w = w +   max(1-r,0)**4
  w = w -5 *max(3/5-r,0)**4  
  w = w +10*max(1/5-r,0)**4
  
  w = w *C_qua_spline_2D/(h*h)
  
  return w  

def Wqui_2Ds(r,h):
  '''
  Normalized 2D quintic kernel : serial version 
  
  h : the smoothing scale (2 sigma)
  '''
  
  # h becomes now H, the kernel-support radius
  h = h*Hh_qui_spline_2D  
  
  # scale the radius
  r = r/h
  
  # normalized kernel part
  
  w = 0
  w = w +   max(1-r,0)**5
  w = w -6 *max(2/3-r,0)**5  
  w = w +15*max(1/3-r,0)**5
  
  w = w *C_qui_spline_2D/(h*h)
  
  return w    


def Wcub_3Ds(r,h):
  '''
  Normalized 3D cubic kernel : serial version 
  
  h : the smoothing scale (2 sigma)
  '''
  
  # h becomes now H, the kernel-support radius
  h = h*Hh_cub_spline_3D  
  
  # scale the radius
  r = r/h
  
  # normalized kernel part
  
  w = 0
  w = w +  max(1-r,0)**3
  w = w -4*max(0.5-r,0)**3  
  
  w = w *C_cub_spline_3D/(h*h*h)
  
  return w


def Wqua_3Ds(r,h):
  '''
  Normalized 3D quadratic kernel : serial version 
  
  h : the smoothing scale (2 sigma)
  '''
  
  # h becomes now H, the kernel-support radius
  h = h*Hh_qua_spline_3D  
  
  # scale the radius
  r = r/h
  
  # normalized kernel part
  
  w = 0
  w = w +   max(1-r,0)**4
  w = w -5 *max(3/5-r,0)**4  
  w = w +10*max(1/5-r,0)**4
  
  w = w *C_qua_spline_3D/(h*h*h)
  
  return w  

def Wqui_3Ds(r,h):
  '''
  Normalized 3D quintic kernel : serial version 
  
  h : the smoothing scale (2 sigma)
  '''
  
  # h becomes now H, the kernel-support radius
  h = h*Hh_qui_spline_3D  
  
  # scale the radius
  r = r/h
  
  # normalized kernel part
  
  w = 0
  w = w +   max(1-r,0)**5
  w = w -6 *max(2/3-r,0)**5  
  w = w +15*max(1/3-r,0)**5
  
  w = w *C_qui_spline_3D/(h*h*h)
  
  return w   



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      param.py
#  brief:     Deal with parameter files
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

try:
    import tkinter as tk
    is_tk = True
except ImportError:
    is_tk = False


import string
import os
import sys
import ast
import numpy as np


def write_ascii_value(value, tpe, name):
    """
    from a name and type and value, return
    an ascii representation of the object
    """

    if tpe == 'ArrayObs':
        if value is not None:
            m = np.ravel(value)
            string = '[%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f]' % (
                m[0], m[1], m[2], m[3], m[4], m[5], m[6], m[7], m[8], m[9], m[10], m[11])
        else:
            string = str(value)
    else:
        string = str(value)

    return string


def read_ascii_value(value, tpe, name):
    """
    from a name and type, return
    an object corresponding to the value given
    """

    if value == "None":
        return None

    if tpe == 'Int':
        try:
            value = int(value)
        except BaseException:
            value = ""

    elif tpe == 'Float':
        try:
            value = float(value)
        except BaseException:
            value = ""

    elif tpe == 'String':
        pass

    elif tpe == 'Tuple':
        try:
            value = ast.literal_eval(value)
        except BaseException:
            value = 0

    elif tpe == 'List':
        try:
            value = ast.literal_eval(value)
        except BaseException:
            value = 0

    elif tpe == 'ArrayObs':
        try:
            value = ast.literal_eval(value)
            value = np.array(value)
            value.shape = (4, 3)
        except BaseException:
            value = None
    else:
        return None

    # check the type
    if isinstance(value, gettype(tpe)) or value is None:
        return value
    else:
        print(gettype(tpe), type(value), value, tpe)
        print(("wrong type returning None :", value, tpe, name))
        return None


def gettype(thetype):

    import types

    if thetype == 'Int':
        return int

    elif thetype == 'Float':
        return float

    elif thetype == 'String':
        return str

    elif thetype == 'Tuple':
        return tuple

    elif thetype == 'List':
        return list

    elif thetype == 'ArrayObs':
        return type(np.ones(1))

    else:
        return None


class Params:
    """
      params = [['imwidth','image width','i',512],
                ['imheight','image height','i',384],
                ['winwidth','window width','f',50.],
                ['winheight','window height','f',37.5]]
    """

    def __init__(self, filename, master):

        self.master = master
        self.params = []
        self.filename = filename

        self.read()

    ###########################

    def set(self, name, value):
        ###########################
        """
        set the value of a parameter
        """

        for param in self.params:
            if param[0] == name:
                param[3] = value

    ###########################

    def get(self, name):
        ###########################
        """
        return the value of a parameter
        """

        for param in self.params:
            if param[0] == name:
                return param[3]

    ###########################
    def get_type(self, name):
        ###########################
        """
        return the type of a parameter
        """

        for param in self.params:
            if param[0] == name:
                return param[2]

    ###########################
    def get_string(self, name):
        ###########################
        """
        return the value of a parameter in a string
        """

        for param in self.params:
            if param[0] == name:
                return write_ascii_value(param[3], param[2], param[0])

    ###########################
    def lists(self):
        ###########################
        """
        print the list of the content of the class
        """

        print((100 * "-"))
        print(("%30s   %30s   %15s (%s)" %
               ("name", "meaning", "value", "type")))
        print((100 * "-"))

        for param in self.params:

            name = param[0]
            nikname = param[1]
            type = param[2]
            value = param[3]

            print(("%30s : %30s = %15s (%s)" % (name, nikname, value, type)))

    ###########################
    def get_dic(self):
        ###########################
        """
        return values of parameters in a dictionary
        """

        dict = {}

        for param in self.params:

            name = param[0]
            value = param[3]

            dict[name] = value

        return dict

    ###########################
    def save(self, filename=None):
        ###########################
        """ ['cd', 'cd', 'Float', 0.0] """

        if filename is None:
            filename = self.filename

#    if not os.path.exists(filename):
#      print "%s does not exists"%(filename)
#      sys.exit()

        f = open(filename, 'w')

        for param in self.params:
            f.write("%s\n" % (param[0]))
            f.write("%s\n" % (param[1]))
            f.write("%s\n" % (param[2]))
            f.write("%s\n" % (self.get_string(param[0])))
            f.write("\n")
        f.close()

    ###########################
    def read(self, filename=None):
        ###########################

        self.params = []

        if filename is None:
            filename = self.filename

        if not os.path.exists(filename):
            print(("%s does not exists" % (filename)))
            sys.exit()

        f = open(filename, 'r')

        while True:
            newparam = []
            line = f.readline()
            if line == '':
                break

            newparam.append(str.strip(line))

            line = f.readline()
            newparam.append(str.strip(line))

            line = f.readline()
            newparam.append(str.strip(line))

            line = f.readline()
            newparam.append(str.strip(line))

            line = f.readline()

            self.params.append(newparam)

        f.close()

        # asign good value

        for param in self.params:
            var = read_ascii_value(param[3], param[2], param[0])
            param[3] = var

    ###########################
    def edit(self):
        ###########################

        Form = ParamForm(self, self.master)


###########################
class ParamForm:
    ###########################

    def __init__(self, params, master=None):

        if is_tk:

            self.root = tk.Toplevel()
            self.master = master

            box = tk.Frame(self.root)
            box.grid(column=0, row=0, sticky=tk.E + tk.N + tk.S + tk.W)

            self.content = {}
            # create all entries
            n = 0

            for param in params.params:
                name = param[0]
                self.content[name] = entryCl(box, n, param, params)
                n = n + 1

            but = tk.Frame(self.root)
            but.grid(column=0, row=1, sticky=tk.E + tk.N + tk.S + tk.W)

            tk.Button(
                but,
                text='Cancel',
                command=self.onCancel).pack(
                side=tk.RIGHT)
            tk.Button(
                but,
                text='Ok',
                command=self.onSubmit).pack(
                side=tk.RIGHT)
            tk.Button(
                but,
                text='Send',
                command=self.onSend).pack(
                side=tk.RIGHT)

        else:
            print("Class ParamForm : tk is not present")

    def onSubmit(self):

        lst = list(self.content.keys())
        for entry in lst:
            self.content[entry].get()

        self.master.send_param()
        self.master.redisplay()
        self.root.destroy()

    def onSend(self):

        lst = list(self.content.keys())
        for entry in lst:
            self.content[entry].get()

        self.master.send_param()
        self.master.redisplay()

    def onCancel(self):
        self.root.destroy()


###########################
class entryCl:
    ###########################

    def __init__(self, master, n, param, params):

        self.param = param
        self.params = params

        name = "%s(%s)" % (param[1], param[2])

        value = write_ascii_value(param[3], param[2], param[0])

        tk.Label(
            master,
            text=name).grid(
            column=0,
            row=n,
            sticky=tk.N +
            tk.S +
            tk.E)
        self.entry = tk.Entry(master, width=20)
        self.entry.insert(tk.INSERT, value)
        self.entry.grid(column=1, row=n, sticky=tk.E + tk.N + tk.S + tk.W)

    def get(self):

        # asign value
        var = read_ascii_value(self.entry.get(), self.param[2], self.param[0])
        # here, as self.params is not a copy, the values are changed in the
        # main program
        self.params.set(self.param[0], var)



###########################################################################################
#  package:   pNbody
#  file:      swift.py
#  brief:     SWIFT file format
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>, Loic Hausammann <loic_hausammann@hotmail.com>
#
# This file is part of pNbody.
###########################################################################################


##########################################################################
#
# GEAR HDF5 CLASS
#
##########################################################################

import numpy as np
import h5py

import pNbody
from pNbody import mpi, error, units

try:				# all this is useful to read files
    from mpi4py import MPI
except BaseException:
    MPI = None



class Nbody_gh5:
  
  
    def get_default_arrays_props(self):
      '''
      get the default properties of arrays considered
      
      "h5name"  :     name in the hdf5 file
      "dtype"   :     numpy dtype                       
      "dim"     :     dimention 
      "ptypes"  :     type of particles that store this property (useless for the moment)
      "read"    :     read it by default or not
      "write"   :     write it by default or not
      "default" :     default values of the components
      "loaded"  :     is the array currently loaded    
      
      # position must always be first, this is a convention
      '''
      
      all_ptypes = list(range(self.get_mxntpe()))
      
      aprops = {}
      
      aprops["pos"] =  {
                        "h5name"  :     "Coordinates", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     3, 
                        "ptypes"  :     all_ptypes,
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["vel"] =  {
                        "h5name"  :     "Velocities", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     3,
                        "ptypes"  :     all_ptypes, 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }
                      
      aprops["mass"] =  {
                        "h5name"  :     "Masses", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     all_ptypes, 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     1,
                        "loaded"  :     False
                      }

      aprops["num"] =  {
                        "h5name"  :     "ParticleIDs", 
                        "dtype"   :     np.uint32,                        
                        "dim"     :     1,
                        "ptypes"  :     all_ptypes, 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }
                          
      aprops["pot"] =  {
                        "h5name"  :     "Potentials", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     all_ptypes, 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["acc"] =  {
                        "h5name"  :     "Acceleration", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     3,
                        "ptypes"  :     all_ptypes, 
                        "read"    :     False,
                        "write"   :     False, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["rsp"] =  {
                        "h5name"  :     "SmoothingLengths", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [0], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["rsp_init"] =  {
                        "h5name"  :     "SmoothingLength", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [0], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }


      aprops["u"] =  {
                        "h5name"  :     "InternalEnergies", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [0], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["u_init"] =  {
                        "h5name"  :     "InternalEnergy", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [0], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["rho"] =  {
                        "h5name"  :     "Densities", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [0], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }
                      





      aprops["snii_thermal_time"] =  {
                        "h5name"  :     "SNII_ThermalTime", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [0], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["snia_thermal_time"] =  {
                        "h5name"  :     "SNIa_ThermalTime", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [0], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }


      aprops["softening"] =  {
                        "h5name"  :     "Softenings", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [1,2], 
                        "read"    :     False,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["minit"] =  {
                        "h5name"  :     "BirthMasses", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [4], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["tstar"] =  {
                        "h5name"  :     "BirthScaleFactors", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [4], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["tstar"] =  {
                        "h5name"  :     "StellarFormationTime",         # same as "BirthScaleFactors"
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [4], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }


      aprops["rhostar"] =  {
                        "h5name"  :     "BirthDensities", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [4], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }



      aprops["sp_type"] =  {
                        "h5name"  :     "StellarParticleType", 
                        "dtype"   :     np.int32,                        
                        "dim"     :     1,
                        "ptypes"  :     [4], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }


      aprops["idp"] =  {
                        "h5name"  :     "ProgenitorIDs", 
                        "dtype"   :     np.uint32,                        
                        "dim"     :     1,
                        "ptypes"  :     [4], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }



      aprops["age"] =  {
                        "h5name"  :     "StellarAge", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [4], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }
                      
      aprops["metals"] =  {
                        "h5name"  :     "Metallicity", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     15,
                        "ptypes"  :     [4], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }

      aprops["mh"] =  {
                        "h5name"  :     "MH", 
                        "dtype"   :     np.float32,                        
                        "dim"     :     1,
                        "ptypes"  :     [4], 
                        "read"    :     True,
                        "write"   :     True, 
                        "default" :     0,
                        "loaded"  :     False
                      }                   


      return aprops  
    
    
    
    def array_is_loaded(self,name):
      return self.arrays_props[name]["loaded"]

    def array_set_loaded(self,name):
      self.arrays_props[name]["loaded"]  = True     
  
    def array_set_unloaded(self,name):
      self.arrays_props[name]["loaded"]  = False 
      
    def array_h5_key(self,name):
      return self.arrays_props[name]["h5name"]
      
    def array_default_value(self,name):
      return self.arrays_props[name]["default"]      
      
    def array_dtype(self,name):
      return self.arrays_props[name]["dtype"]      

    def array_dimension(self,name):
      return self.arrays_props[name]["dim"]   

    def array_read(self,name):
      return self.arrays_props[name]["read"] 
        

  
    def get_npart_from_dataset(self,ptypes=None):
      '''
      Get npart from the block storing the positions
      '''

      if ptypes is None:
        ptypes = list(range(self.get_mxntpe()))
        
      h5_key = self.array_h5_key('pos')  
      
      if mpi.mpi_NTask() > 1:
          fd = h5py.File(self.p_name_global[0],'r',driver="mpio",comm=MPI.COMM_WORLD)
      else:
          fd = h5py.File(self.p_name_global[0],'r')
      
      npart = np.zeros(self.get_mxntpe(),np.uint32)
      
      # loop over particle types
      for i_type in ptypes:
        PartType = "PartType%d" % i_type
        
        if PartType in fd:
          block = fd[PartType]
          
          # if the dataset is present
          if h5_key in fd[PartType].keys():
            
            data_lenght   = block[h5_key].len()
          
            idx0=0
            idx1=data_lenght
            if mpi.mpi_NTask() > 1:
              idx0, idx1 = self.get_particles_limits(data_lenght)
            
            npart[i_type] = idx1-idx0   
            
          else:
            raise error.FormatError("%s is not in the file."%(h5_key))     
      
      
      # close the file
      mpi.mpi_barrier()
      fd.close()
      
      return npart
      

    def set_readable_arrays(self,ptypes=None):
      '''
      Set the arrays to be readable or not according to their presence in the blocks
      '''

      if ptypes is None:
        ptypes = list(range(self.get_mxntpe()))
        
      
      if mpi.mpi_NTask() > 1:
          fd = h5py.File(self.p_name_global[0],'r',driver="mpio",comm=MPI.COMM_WORLD)
      else:
          fd = h5py.File(self.p_name_global[0],'r')
      

      # loop over all arrays flagged as read=True in self.arrays_props
      for name in self.arrays_props.keys():
        
        # skip array if a list of array is provided in self.array
        if self.arrays is not None:
          if not name in self.arrays:
            self.arrays_props[name]["read"]=False
        
        
        if self.array_read(name):

          h5_key        = self.array_h5_key(name)
          flag = 0

          # loop over particle types
          for i_type in ptypes:
            PartType = "PartType%d" % i_type
            
            # if the dataset is present
            if PartType in fd:
              if h5_key in fd[PartType].keys():   
                flag = 1
            
          # if no dataset is found, 
          if flag==0:
            self.arrays_props[name]["read"]=False
          
      
      fd.close()  
        
      
  
    def load(self,name=None,ptypes=None,force=False):
      '''
      Load array from the hdf5 file.
      This function relays on the info stored self.arrays_props
      Here, we assume that npart is known and correct
      '''
      
      if ptypes is None:
        ptypes = list(range(self.get_mxntpe()))
      
      if name not in self.arrays_props:
        self.warning("load error : %s not defined in self.arrays_props"%(name))
        return
      
      dtype         = self.array_dtype(name)
      h5_key        = self.array_h5_key(name)
      default_value = self.array_default_value(name)
      dim           = self.array_dimension(name)
      
      if mpi.mpi_NTask() > 1:
          fd = h5py.File(self.p_name_global[0],'r',driver="mpio",comm=MPI.COMM_WORLD)
      else:
          fd = h5py.File(self.p_name_global[0],'r')
      
      

            
      if not self.array_is_loaded(name) or force:
        self.message("loading %s"%name)
      

        # erase existing instance
        try: 
          delattr(self,name)
        except AttributeError:
          pass

        
        # loop over particle types
        for i_type in ptypes:
          PartType = "PartType%d" % i_type
          
          if PartType in fd:
            block = fd[PartType]
            
            # if the dataset is present
            if h5_key in fd[PartType].keys():
              
              data_lenght   = block[h5_key].len()
                          
              idx0=0
              idx1=data_lenght
              if mpi.mpi_NTask() > 1:
                idx0, idx1 = self.get_particles_limits(data_lenght)
                
              # sanity check
              if idx1-idx0 !=self.npart[i_type]:
                raise error.FormatError("npart[%d]=%d is not equal to idx1-idx0=%d"%(i_type,npart[i_type],idx1-idx0))                 
                
              # read             
              data = block[h5_key][idx0:idx1].astype(dtype)
              
            
            # if the dataset is not present, compute from default value
            else:
              if dim==1:
                data = (default_value * np.ones(self.npart[i_type])).astype(dtype)
              else:
                data = (default_value * np.ones((self.npart[i_type],dim))).astype(dtype)
           
            try:
              setattr(self, name, np.concatenate((getattr(self, name), data)))
            except AttributeError:
              setattr(self, name, data)
          
    
      
      
      # set it as loaded        
      self.array_set_loaded(name)

      # close the file 
      mpi.mpi_barrier()
      fd.close()
        

        
        
        
    def read_arrays(self,ptypes=None):    
      '''
      Read arrays that must be loaded.
      '''      
      
      if ptypes is None:
        ptypes = list(range(self.get_mxntpe()))      
      
      # loop over all arrays flagged as read=True in self.arrays_props
      for name in self.arrays_props.keys():
        if self.array_read(name) and not self.array_is_loaded(name):
          self.load(name,ptypes)
      
 
 

    def _init_spec(self):
        # create an empty header (in case the user is creating a dataset)
        self._header = []

    def get_excluded_extension(self):
        """
        Return a list of file to avoid when extending the default class.
        """
        return []

    def getParticleMatchingDict(self):
        """
        Return a list of file to avoid when extending the default class.
        """
        
        index = {
            'gas':    0,
            'halo':   1,
            'disk':   2,
            'sink':   3,
            'stars':  4,
            'bndry':  2,
            'stars1': 4,
            'halo1':  1}
        
        return index 

    def check_spec_ftype(self):
        try:
            if mpi.NTask > 1:
                fd = h5py.File(self.p_name_global[0],'r',driver="mpio",comm=MPI.COMM_WORLD)
            else:
                fd = h5py.File(self.p_name_global[0])
            
            
            test1 = False
            test2 = False
            
            header = fd["Header"]
            if not "Compactify_Version" in header.attrs:
              test1 = True

            #test1 = "Units" not in fd
            #test1 = test1 or "Unit temperature in cgs (U_T)" not in fd["Units"].attrs
            #test2 = "PartType0/Offset" in fd
            
            
            fd.close()
            if test1 or test2:
                raise error.FormatError("gizmo")

        except IOError as e:
            self.warning("gizmo not recognized: %s" % e)
            raise error.FormatError("gizmo")

    def set_pio(self, pio):
        """ Overwrite function
        """
        pNbody.Nbody.set_pio(self, "no")

    def get_read_fcts(self):
        return [self.read_particles]

    def get_write_fcts(self):
        return [self.write_particles]

    def get_mxntpe(self):
        return 6

    def get_header_translation(self):
        """
        Gives a dictionnary containing all the header translation.
        If a new variable is possible in the HDF5 format, only the translation
        is required for the reader/writer.
        As h5py is not supporting dictionnary, they need special care when reading/writing.
        """
        # dict containing all the main header variables (=> easy acces)
        # e.g. self.npart will contain NumPart_ThisFile
        header_var = {}

        # Size variables
        header_var["Header/NumPart_ThisFile"] = "npart"
        header_var["Header/NumPart_Total"] =  "npart" #   "npart_tot" check what must be done properly here
        header_var["Header/NumPart_Total_HighWord"] = "nallhw"
        header_var["Header/MassTable"] = "massarr"
        header_var["Header/NumFilesPerSnapshot"] = "num_files"
        header_var["Header/BoxSize"] = "boxsize"
        header_var["Header/Flag_Entropy_ICs"] = "flag_entr_ics"
        header_var["Header/Compactify_Version"] = "version"
        
        

        # Physics
        #header_var["Header/Scale-factor"] = "atime"
        #header_var["Header/Time"] = "time"
        header_var["Header/Redshift"] = "redshift"
        header_var["Header/Time"]     = "atime"
        header_var["Header/Redshift"] = "redshift"
         
        

        # Cosmology
        #header_var["Cosmology/Omega_b"] = "omegab"
        header_var["Header/Omega0"]      = "omega0"
        header_var["Header/OmegaLambda"] = "omegalambda"
        header_var["Header/HubbleParam"] = "hubbleparam"
        header_var["Header/Cosmorun"] = "cosmorun"        


        # Units
        #header_var["Units/Unit velocity in cgs (U_V)"] = "UnitVelocity_in_cm_per_s"
        #header_var["Units/Unit length in cgs (U_L)"] = "UnitLength_in_cm"
        #header_var["Units/Unit mass in cgs (U_M)"] = "UnitMass_in_g"
        #header_var["Units/Unit time in cgs (U_t)"] = "Unit_time_in_cgs"
        #header_var["Units/Unit temperature in cgs (U_T)"] = "Unit_temp_in_cgs"
        #header_var["Units/Unit current in cgs (U_I)"] = "Unit_current_in_cgs"



        return header_var

    def get_list_header(self):
        """
        Gives a list of header directory from self.get_header_translation
        """
        list_header = []
        trans = self.get_header_translation()
        for key, tmp in list(trans.items()):
            directory = key.split("/")[0]
            if directory not in list_header:
                list_header.append(directory)
        return list_header

    def get_array_translation(self):
        """
        Gives a dictionnary containing all the header translation with the particles type
        requiring the array.
        If a new variable is possible in the HDF5 format, the translation,
        default value (function get_array_default_values) and dimension (function get_array_dimension)
        are required for the reader/writer.
        """
        # hdf5 names -> pNbody names
        # [name, partType, type] where partType is a list of particles with the datas
        # True means all, False means none
        # and type is the data type
        ntab = {}
        
        ###################
        # common data
        ntab["Acceleration"] = ["acc", True, np.float32]
        ntab["Coordinates"] = ["pos", True, np.float32]
        ntab["Velocities"] = ["vel", True, np.float32]
        ntab["ParticleIDs"] = ["num", True, np.uint32]
        ntab["Masses"] = ["mass", True, np.float32]
        
        
        ntab["Potentials"] = ["pot", True, np.float32]
        
        ###################
        # gas data
        ntab["SmoothingLength"] = ["rsp_init", [0], np.float32]       # needed for ICs...
        ntab["InternalEnergy"] = ["u_init", [0], np.float32]          # needed for ICs...

        ntab["SmoothingLengths"] = ["rsp", [0], np.float32]
        ntab["InternalEnergies"] = ["u", [0], np.float32]
        ntab["Densities"] = ["rho", [0], np.float32]
        
        #ntab["Pressure"] = ["p", [0], np.float32]
        #ntab["Entropy"] = ["a", [0], np.float32]
        

                
        ###################
        # dm data
        ntab["Softenings"]        = ["softening",   [1], np.float32]   
        
        ###################
        # bndry data
        ntab["Softenings"]        = ["softening",   [2], np.float32]             
        
        
        ###################
        # stars data
        ntab["BirthMasses"]          = ["minit",   [4], np.float32]
        ntab["BirthScaleFactors"]    = ["tstar",   [4], np.float32]        
        ntab["StellarFormationTime"] = ["tstar",   [4], np.float32]
        ntab["BirthDensities"]       = ["rhostar", [4], np.float32]
        ntab["MetalMassFractions"]   = ["metals", [4], np.float32]
        #ntab["SmoothedMetalMassFractions"] = ["metals", [4], np.float32]
        ntab["ProgenitorIDs"]        = ["idp", [4], np.uint32]
        
        ntab["StellarAge"]           = ["age",   [4], np.float32]
        ntab["MH"]                   = ["mh",    [4], np.float32]
         
        ntab["SmoothingLengths"] = ["rsp",    [4], np.float32]

        
        ntab["MagVIS"] = ["MagVIS",    [4], np.float32]
        ntab["MagY"] = ["MagY",    [4], np.float32]
        ntab["MagJ"] = ["MagJ",    [4], np.float32]
        
        
        return ntab

    def get_array_default_value(self):
        """
        Gives a dictionary of default value for pNbody's arrays
        """
        # default value
        dval = {}
        dval["p"] = 0.0
        dval["a"] = 0.0
        dval["acc"] = 0.0
        dval["pos"] = 0.0
        dval["vel"] = 0.0
        dval["num"] = 0.0
        dval["mass"] = 0.0
        dval["pot"] = 0.0
        dval["u"] = 0.0
        dval["rho"] = 0.0
        dval["metals"] = 0.0
        dval["opt1"] = 0.0
        dval["opt2"] = 0.0
        dval["rsp"] = 0.0
        dval["softening"] = 0.0
        dval["minit"] = 0.0
        dval["tstar"] = 0.0
        dval["rhostar"] = 0.0
        dval["idp"] = 0.0
        dval["rsp_stars"] = 0.0
        dval["rho_stars"] = 0.0
        dval["snii_thermal_time"] = 0.0
        dval["snia_thermal_time"] = 0.0
        dval["entropy"] = 0.0
        dval["p"] = 0.0
        
        dval["MagVIS"] = 0.0
        dval["MagY"] = 0.0
        dval["MagJ"] = 0.0 
               
        return dval

    def get_array_dimension(self):
        """
        Gives a dictionary of dimension for pNbody's arrays
        """
        # dimension
        vdim = {}
        vdim["pos"] = 3
        vdim["vel"] = 3

        vdim["p"] = 1
        vdim["a"] = 1
        vdim["acc"] = 3
        vdim["num"] = 1
        vdim["mass"] = 1
        vdim["pot"] = 1
        vdim["u"] = 1
        vdim["rho"] = 1
        vdim["opt1"] = 1
        vdim["opt2"] = 1
        vdim["rsp"] = 1
        vdim["softening"] = 1
        vdim["minit"] = 1
        vdim["tstar"] = 1
        vdim["rhostar"] = 1
        vdim["idp"] = 1
        vdim["rsp_stars"] = 1
        vdim["rho_stars"] = 1
        vdim["snii_thermal_time"] = 1
        vdim["snia_thermal_time"] = 1
        vdim["entropy"] = 1
        vdim["p"] = 1
        vdim["metals"] = self.ChimieNelements
        
        vdim["MagVIS"] = 1
        vdim["MagY"]   = 1
        vdim["MagJ"]   = 1 
                
        return vdim

    def get_default_spec_vars(self):
        """
        return specific variables default values for the class
        """

        return {'massarr': np.array([0, 0, 0, 0, 0, 0]),
                'atime': 0.,
                'redshift': 0.,
                'flag_sfr': 0,
                'flag_feedback': 0,
                'npart_tot': np.array([0, 0, self.nbody, 0, 0, 0]),
                'npart': np.array([0, 0, self.nbody, 0, 0, 0]),
                'flag_cooling': 0,
                'num_files': 1,
                'boxsize': 0.,
                'omega0': 0.,
                'omegalambda': 0.,
                'hubbleparam': 0.,
                'flag_age': 0.,
                'flag_metals': 0.,
                'nallhw': np.array([0, 0, 0, 0, 0, 0]),
                'flag_entr_ics': 0,
                'flag_chimie_extraheader': 0,
                'critical_energy_spec': 0.,
                'empty': 48 * '',
                'comovingintegration': True,
                'hubblefactorcorrection': False,
                'comovingtoproperconversion': True,
                'ChimieNelements': 0,
                'utype':"swift",
                'cosmorun': 1,
                'version': 0
                }

    def get_particles_limits_from_npart(self, i):
        """ Gives the limits for a thread.
        In order to get the particles, slice them like this pos[start:end].
        :param int i: Particle type
        :returns: (start, end)
        """
        nber = float(self.npart_tot[i]) / mpi.mpi_NTask()
        start = int(mpi.mpi_ThisTask() * nber)
        end = int((mpi.mpi_ThisTask() + 1) * nber)
        return start, end
        
    def get_particles_limits(self,size):
        """ Gives the limits for a thread.
        In order to get the particles, slice them like this pos[start:end].
        :param int i: Particle type
        :returns: (start, end)
        """
        nber = float(size) / mpi.mpi_NTask()
        start = int(mpi.mpi_ThisTask() * nber)
        end = int((mpi.mpi_ThisTask() + 1) * nber)
        return start, end        
        

    def set_local_value(self):
        N = mpi.mpi_NTask()
        if N == 1:
            return
        part = len(self.npart_tot)
        for i in range(part):
            s, e = self.get_particles_limits_from_npart(i)
            self.npart[i] = e - s

    def get_massarr_and_nzero(self):
        """
        return massarr and nzero

        !!! when used in //, if only a proc has a star particle,
        !!! nzero is set to 1 for all cpu, while massarr has a length of zero !!!
        """

        if self.has_var('massarr') and self.has_var('nzero'):
            if self.massarr is not None and self.nzero is not None:
                self.warning("warning : get_massarr_and_nzero : here we use massarr and nzero %s %s"%(self.massarr,self.nzero))
                return self.massarr, self.nzero

        massarr = np.zeros(len(self.npart), float)
        nzero = 0

        # for each particle type, see if masses are equal
        for i in range(len(self.npart)):
            first_elt = sum((np.arange(len(self.npart)) < i) * self.npart)
            last_elt = first_elt + self.npart[i]

            if first_elt != last_elt:
                c = (self.mass[first_elt] ==
                     self.mass[first_elt:last_elt]).astype(int)
                if sum(c) == len(c):
                    massarr[i] = self.mass[first_elt]
                else:
                    nzero = nzero + len(c)

        return massarr.tolist(), nzero

    def read_particles(self, f):
        """
        read gadget file
        """
                
        from copy import deepcopy
        import time
                
        # go to the end of the file
        if f is not None:
            f.seek(0, 2)

        if mpi.mpi_NTask() > 1:
            fd = h5py.File(self.p_name_global[0],'r',driver="mpio",comm=MPI.COMM_WORLD)
        else:
            fd = h5py.File(self.p_name_global[0],'r')

        ################
        # read header
        ################
        self.message("reading header...")

        # set default values
        default = self.get_default_spec_vars()
        
        
        for key, i in list(default.items()):
            setattr(self, key, i)

        # get values from snapshot
        trans = self.get_header_translation()

        list_header = self.get_list_header()
        

        for name in list_header:
            if name not in fd:
                continue
                
            # e.g. create self.npart with the value
            # fd["Header"].attrs["NumPart_ThisFile"]
                        
            for key in fd[name].attrs:
                                                                    
              
                full_name = name + "/" + key
                if full_name not in list(trans.keys()):
                    trans[full_name] = full_name

                tmp = fd[name].attrs[key]
                if isinstance(tmp, bytes) and tmp == "None":
                    tmp = None
                if isinstance(tmp, bytes):
                    tmp = tmp.decode('utf-8')
                setattr(self, trans[full_name], tmp)
                
                self.message("%s %s >> %s %s"%(name,key,trans[full_name],tmp),verbosity=2)
                
        

        # additional stuffs
        # we assume its a cosmo run
        
        
        # additional stuffs
        if self.has_var('cosmorun'):
          if self.cosmorun==1:
            self.cosmorun=1
            self.HubbleFactorCorrectionOn()
            self.ComovingToProperConversionOn()
          else:
            self.cosmorun=0
            self.HubbleFactorCorrectionOff()
            self.ComovingToProperConversionOff()            
        else:   
          self.cosmorun=1
          self.HubbleFactorCorrectionOn()
          self.ComovingToProperConversionOn()                 


        
        # get local value from global ones
        self.set_local_value()
        
       

        
        
        
        ###############
        # read units
        ###############
        self.message("reading units...")

        # define system of units
        params = {}
        params['UnitLength_in_cm']         = 3.085678e21
        params['UnitVelocity_in_cm_per_s'] = 1.0e5
        params['UnitMass_in_g']            = 1.989e43
        self.localsystem_of_units = units.Set_SystemUnits_From_Params(params)
        
        mpi.mpi_barrier()
        fd.close()
        
        ################
        # read particles
        ################

                
        # get npart from the blocks
        self.npart = self.get_npart_from_dataset(ptypes=self.ptypes)
        
        # check that arrays are present
        self.set_readable_arrays(ptypes=self.ptypes)
        
        # load all arrays according to the information
        # provided by self.arrays_props
        self.read_arrays(ptypes=self.ptypes)
                

        # set tpe
        self.tpe = np.array([], np.int32)
        for i in range(len(self.npart)):
            self.tpe = np.concatenate((self.tpe, np.ones(self.npart[i]) * i))

        # compute nzero
        nzero = 0
        mass = np.array([])

        for i in range(len(self.npart)):
            if self.massarr[i] == 0:
                nzero = nzero + self.npart[i]
            else:
                self.warning("Massarr is not supported! Please specify the mass of all the particles!",verbosity=2)

        self.nzero = nzero

        mpi.mpi_barrier()
        
        


        # specific final conversions
        

          
          


    def write_particles(self, f):
        """
        specific format for particle file
        """
        # go to the end of the file
        if f is not None:
            f.seek(0, 2)
            

        name = "Unit_temp_in_cgs"
        if not hasattr(self, name):
            setattr(self, name, 1.0)

        name = "Unit_current_in_cgs"
        if not hasattr(self, name):
            setattr(self, name, 1.0)

        import h5py
        # not clean, but work around pNbody
        filename = self.p_name_global[0]
        # open file
        if mpi.mpi_NTask() > 1:
            from mpi4py import MPI
            h5f = h5py.File(filename, "w", driver="mpio", comm=MPI.COMM_WORLD)
        else:
            h5f = h5py.File(filename, "w")

        # add units to the usual gh5 struct
        if hasattr(self, "unitsparameters"):
            units = self.unitsparameters.get_dic()
            for key, i in list(units.items()):
                if not hasattr(self, key):
                    setattr(self, key, i)

        if hasattr(self,"UnitVelocity_in_cm_per_s") and hasattr(self,"UnitLength_in_cm"):
          self.Unit_time_in_cgs = self.UnitLength_in_cm / self.UnitVelocity_in_cm_per_s
        
 
          

        ############
        # HEADER
        ############
        self.message("Writing header...")

        list_header = self.get_list_header()
        trans = self.get_header_translation()
        # cheat a little bit in order to get the real number of particles in
        # this file
        trans["Header/NumPart_ThisFile"] = "npart_tot"
        
        for name in list_header:
            h5f.create_group(name)
        
                    
            
        for key in self.get_list_of_vars():
          
            if key in list(trans.values()):
                ind = list(trans.values()).index(key)
                name, hdf5 = list(trans.keys())[ind].split("/")
                value = getattr(self, key)
                
            
                if type(value) is not str:
                  value = np.array(value)
                else:
                  value = np.array(value,dtype="S")  
                  
                if value.shape == ():
                  value = np.array([value])
                  
                
                if not isinstance(value, dict):
                    if value is None:
                        h5f[name].attrs[hdf5] = "None"
                    else:
                        h5f[name].attrs[hdf5] = value

        ##############
        # PARTICULES
        ##############
        self.message("Writing particles...")

        for i in range(len(self.npart)):
            if self.massarr[i] != 0:
              self.warning("Massarr is not supported! Please specify the mass of all the particles!",verbose=2)

        ntab = self.get_array_translation()
        
                
        # get particles type present
        type_part = []
        for i in range(len(self.npart)):
            if self.npart[i] > 0:
                type_part.append(i)

        # write particles
        for i in type_part:
            if mpi.mpi_NTask() > 1:
                init, end = self.get_particles_limits_from_npart(i)                
            self.message("Writing particles (type %i)..." % i,verbose=2)
            group = "PartType%i" % i
            grp = h5f.create_group(group)
            nb_sel = self.select(i)                                   # this is really bad as we duplicate the model
            
            for key, j in list(ntab.items()):
                varname = j[0]
                var_type = j[1]
                
  
                if (var_type is True) or i in var_type:
                    if hasattr(nb_sel, varname):
                      
                        # get and transform type
                        tmp = getattr(nb_sel, varname)
                        if tmp is None:
                            continue
                        tmp = tmp.astype(j[2])
                        # create dataset
                        size = list(tmp.shape)
                        size[0] = self.npart_tot[i]
                        if mpi.mpi_NTask() > 1:
                            dset = grp.create_dataset(key, size, dtype=j[2])
                            dset[init:end] = tmp
                        else:
                            h5f[group][key] = tmp

        

        h5f.close()
        
        

        



###########################################################################################
#  package:   pNbody
#  file:      dissipatives.py
#  brief:
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################
from Nbody import *
from Numeric import *

from tkinter import *


class Dissip:

    def __init__(self, master):

        self.master = master

        self.varRad = StringVar()

        self.root = Toplevel()
        box = Frame(self.root)
        box.grid(column=0, row=0, sticky=E + N + S + W)

        Label(
            box,
            text='dissipative period %5d' %
            (self.master.X.pdis)).grid(
            columnspan=2,
            column=0,
            row=0,
            sticky=E +
            N +
            S +
            W)
        Label(
            box,
            text='dissipative time   %5d' %
            (self.master.X.tdis)).grid(
            columnspan=2,
            column=0,
            row=1,
            sticky=E +
            N +
            S +
            W)
        Label(
            box,
            text='select particules with').grid(
            columnspan=2,
            column=0,
            row=2,
            sticky=E +
            N +
            S +
            W)

        Label(box, text='tdis < ').grid(column=0, row=3, sticky=E + N + S + W)
        Label(box, text='tdis > ').grid(column=0, row=5, sticky=E + N + S + W)

        self.andRad = Radiobutton(
            box,
            text="and",
            command=self.cmdRad,
            var=self.varRad,
            value="and")
        self.andRad.grid(column=0, row=4)
        self.orRad = Radiobutton(
            box,
            text="or",
            command=self.cmdRad,
            var=self.varRad,
            value="or")
        self.orRad.grid(column=1, row=4)

        self.t1Ent = Entry(box)
        self.t1Ent.grid(column=1, row=3, sticky=E + N + S + W)
        self.t2Ent = Entry(box)
        self.t2Ent.grid(column=1, row=5, sticky=E + N + S + W)

        but = Frame(self.root)
        but.grid(column=0, row=1, sticky=E + N + S + W)
        Button(but, text='Cancel', command=self.onCancel).pack(side=RIGHT)
        Button(but, text='Ok', command=self.onSubmit).pack(side=RIGHT)

    def cmdRad(self):
        self.selectmode = self.varRad.get()

    def onCancel(self):
        self.root.destroy()

    def onSubmit(self):

        t1 = float(self.t1Ent.get())
        t2 = float(self.t2Ent.get())

        # selection

        c1 = less(self.master.X.dis, t1)
        c2 = greater(self.master.X.dis, t2)

        if self.selectmode == 'and':
            c = c1 * c2
        elif self.selectmode == 'or':
            c = 1 - ((1 - c1) * (1 - c2))

        self.master.X = self.master.X.selectc(c)

        self.master.display()
        self.master.display_info(self.master.X)

        self.root.destroy()


Dissip(self)



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      fourier.py
#  brief:     Fourier Transform
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


import numpy as np
from numpy import fft


####################################
def tofrec(fmin, fmax, n, i):
    ####################################
    nn = float(n)
    if i <= n // 2 + 1:
        return (i / nn + 0.5) * (fmax - fmin) + fmin
    else:
        return (i / nn - 0.5) * (fmax - fmin) + fmin


####################################
def fourier(x, y):
    ####################################
    """

    Fct = Sum_(m=1,n) amp_m cos( 2.pi f_m phi + phi_m )
        = Sum_(m=1,n) amp_m cos(        m phi + phi_m )

    m = 2.pi f_m

    """

    dx = (x.max() - x.min()) / (len(x) - 1)
    fmin = -1. / (2. * dx)
    fmax = -fmin

    # f = fromfunction(lambda ii:tofrec(fmin,fmax,len(x),ii) ,(len(x),))
    f = np.array([], float)
    for i in range(len(x)):
        f = np.concatenate((f, np.array([tofrec(fmin, fmax, len(x), i)])))

    ## FFT ##
    ffft = fft.fft(y)

    amp = np.sqrt(ffft.real**2 + ffft.imag**2)  # amplitude du signal d'entree
    phi = np.arctan2(ffft.imag, ffft.real)		# phase a p/2 pres
    phi = np.fmod(phi + 2 * np.pi, 2. * np.pi)  # de 0 a 2*pi (ok avec phase positive)

    amp = 2. * amp / len(x)			            # doubler ok, /len(x) ???

    f = f[0:len(x) // 2]				        # on sort que les frequences positives
    amp = amp[0:len(x) // 2]
    phi = phi[0:len(x) // 2]

    amp[0] = amp[0] / 2.

    return f, amp, phi



#!/usr/bin/env python3

###########################################################################################
#  package:   pNbody
#  file:      libbastitime.py
#  brief:     Basti RGB
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
###########################################################################################

import numpy as np
import scipy
import scipy.interpolate
from .libRGBs import NRGB


def readfile(file):
    return(np.genfromtxt(file))


ages = np.linspace(0.3, 17, 50)
zs = [-3.62, -2.62, -2.14, -1.84, -1.62, -1.31, -
      1.01, -0.70, -0.60, -0.29, -0.09, 0.05, 0.16]


class BastiRGB(NRGB):

    def ReadMS(self):
        self.dataMS = readfile(self.fileMS)

    def ReadRGB(self):
        self.dataRGB = readfile(self.fileRGB)

    def InterpolateMS(self):
        # Interpolate so you get the mass given the values
        self.InterpolateMS = scipy.interpolate.RectBivariateSpline(
            ages, zs, self.dataMS, kx=1, ky=1)

    def InterpolateTRGB(self):

        self.InterpolateTRGB = scipy.interpolate.RectBivariateSpline(
            ages, zs, self.dataRGB, kx=1, ky=1)



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      iofunc.py
#  brief:     Input/Output functions
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


# standard modules
import os
import sys
import string
import types
import pickle
import csv
import io
import yaml
import getpass
import time
  
# array module
import numpy as np
import astropy.io.fits as pyfits
from . import mpiwrapper as mpi
from . import libutil
import pNbody

#################################
def checkfile(name):
    #################################
    """
    Check if a file exists. An error is generated if the file
    does not exists.

    Parameters
    ----------
    name : the path to a filename



    Examples
    --------
    >>> io.checkfile('an_existing_file')
    >>>

    >>> io.checkfile('a_non_existing_file')
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "/home/epfl/revaz/local/lib64/python2.6/site-packages/pNbody/io.py", line 33, in checkfile
        raise IOError(915,'file %s not found ! Pease check the file name.'%(name))
    IOError: [Errno 915] file nofile not found ! Pease check the file name.

    """

    if name is None:
        raise Exception("file name set to None ! Please check the file name.")

    if not os.path.isfile(name):
        raise IOError(
            915,
            'file %s not found ! Pease check the file name.' %
            (name))


#################################
def end_of_file(f, pio='no', MPI=None):
    #################################
    """
    Return True if we have reached the end of the file f, False instead

    Parameters
    ----------
    f : ndarray or matrix object
        an open file
    pio : 'yes' or 'no'
        if the file is read in parallel or not
    MPI : MPI communicator


    Returns
    -------
    status : Bool
             True if the we reached the end of the file
             False if not
    """

    if pio == 'no':

        # here, the master decide for all slaves

        if mpi.ThisTask == 0:

            p1 = f.tell()
            f.seek(0, 2)
            p2 = f.tell()
            f.seek(p1)

            if p1 == p2:
                status = True
            else:
                status = False

        else:
            status = None

        status = mpi.mpi_bcast(status, 0)

        return status

    else:

        # each processus decide for himself

        p1 = f.tell()
        f.seek(0, 2)
        p2 = f.tell()
        f.seek(p1)

        if p1 == p2:
            status = True
        else:
            status = False

        return status

#####################################################


def write_array(file, vec):
    #####################################################
    """
    Write an array to a file, in a very simple ascii format.

    Parameters
    ----------
    file : the path to a file
    vec : an ndarray object


    Examples
    --------
    >>> from numpy import *
    >>> x = array([1,2,3])
    >>> io.write_array('/tmp/array.dat',x)
    """

    f = open(file, 'w')
    for i in range(len(vec)):
        f.write("%f\n" % vec[i])
    f.close()


#####################################################
def read_ascii(
        file,
        columns=None,
        lines=None,
        dtype=np.float32,
        skipheader=False,
        cchar='#'):
    #####################################################
    """
    Read an ascii file.
    The function allows to set the number of columns or line to read.
    If it contains a header, the header is used to label all column. In
    this case, a dictionary is returned.


    Parameters
    ----------
    file : the path to a file or an open file
    columns : list
               the list of the columns to read
               if none, all columns are read
    lines : list
            the list of the lines to read
            if none, all lines are read
    dtype : dtype
            the ndtype of the objects to read
    skipheader : bool
                 if true, do not read the header
                 if there is one
    cchar : char
            lines begining with cchar are skiped
            the first line is considered as the header

    Returns
    -------
    data : Dict or ndarray
           A python dictionary or an ndarray object

    Examples
    --------
    >>> from numpy import *
    >>> x = arange(10)
    >>> y = x*x
    >>> f = open('afile.txt','w')
    >>> f.write("# x y")
    >>> for i in xrange(len(x)):
    ...	f.write('%g %g'%(x[i],y[i]))
    ...
    >>> f.close()
    >>> from pNbody import iofunc as io
    >>> data = io.read_ascii("afile.txt")
    >>> data['x']
    array([ 0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9.])
    >>> data['y']
    array([  0.,   1.,   4.,   9.,  16.,  25.,  36.,  49.,  64.,  81.])
    """

    def RemoveComments(l):
        if l[0] == cchar:
            return None
        else:
            return l

    def toNumList(l):
        return list(map(dtype, l))

    if not isinstance(file, io.TextIOWrapper):
        f = open(file, 'r')
    else:
        f = file

    # read header while there is one
    while True:
        fpos = f.tell()
        header = f.readline()
        if header[0] != cchar:
            f.seek(fpos)
            header = None
            break
        else:
            if skipheader:
                header = None
            else:
                # create dict from header
                header = str.strip(header[2:])
                elts = str.split(header)
                break

    """
  # read header if there is one
  header = f.readline()
  if header[0] != cchar:
    f.seek(0)
    header = None
  else:
    if skipheader:
      header = None
    else:
      # create dict from header
      header = str.strip(header[2:])
      elts = str.split(header)
  """

    # now, read the file content
    lines = f.readlines()

    # remove trailing
    lines = list(map(str.strip, lines))

    # remove comments
    #lines = map(RemoveComments, lines)

    # split
    lines = list(map(str.split, lines))

    # convert into float
    lines = list(map(toNumList, lines))

    # convert into array
    lines = np.array(list(map(np.array, lines)))

    # transpose
    lines = np.transpose(lines)

    if header is not None:
        iobs = {}
        i = 0
        for elt in elts:
            iobs[elt] = i
            i = i + 1

        vals = {}
        for key in list(iobs.keys()):
            vals[key] = lines[iobs[key]]

        return vals

    # return
    if columns is None:
        return lines
    else:
        return lines.take(axis=0, indices=columns)


#####################################################
def write_dmp(file, data):
    #####################################################
    """
    Write a dmp (pickle) file. In other word,
    dump the data object.

    Parameters
    ----------
    file : the path to a file
    data : a pickable python object

    Examples
    --------
    >>> x = {'a':1,'b':2}
    >>> io.write_dmp('/tmp/afile.dmp',x)
    """

    f = open(file, 'wb')
    pickle.dump(data, f)
    f.close()


#####################################################
def read_dmp(file):
    #####################################################
    """
    Read a dmp (pickle) file.

    Parameters
    ----------
    file : the path to a file

    Returns
    -------
    data : a python object

    Examples
    --------
    >>> x = {'a':1,'b':2}
    >>> io.write_dmp('/tmp/afile.dmp',x)
    >>> y = io.read_dmp('/tmp/afile.dmp')
    >>> y
    {'a': 1, 'b': 2}
    """

    f = open(file, 'rb')
    data = pickle.load(f)
    f.close()
    return data


#####################################################
def old_WriteFits(data, filename, extraHeader=None):
    #####################################################
    """
    Write a fits file
    """
    # image creation
    fitsimg = pyfits.HDUList()

    # add data
    hdu = pyfits.PrimaryHDU()
    hdu.data = data
    fitsimg.append(hdu)

    # add keys
    keys = []
    if extraHeader is not None:
        # keys.append(('INSTRUME','st4 SBIG ccd camera','Instrument name'))
        # keys.append(('LOCATION',"175 OFXB St-Luc (VS)",'Location'))
        keys = extraHeader

    hdr = fitsimg[0].header
    for key in keys:
        hdr.update(key[0], key[1], comment=key[2])

    fitsimg.writeto(filename)

#####################################################
def WriteFits(data, filename, extraHeader=None):
#####################################################
  """
  Write a fits file
  """

  from astropy.io import fits
  
  hdu = fits.PrimaryHDU(data)
  hdu.writeto(filename)
  
  

#####################################################
def ReadFits(filename):
    #####################################################
    """
    Read a fits file.
    """
    # read image
    fitsimg = pyfits.open(filename)
    data = fitsimg[0].data
    return data


#################################
def computeBlockSize(shape, data_type):
    #################################
    # compute the number of bytes that should be read
    nbytes_to_read = None
    if shape is not None:
        shape_a = np.array(shape)
        nelts_to_read = shape_a[0]
        for n in shape_a[1:]:
            nelts_to_read = nelts_to_read * n
        nbytes_to_read = nelts_to_read * np.dtype(data_type).itemsize
    return nbytes_to_read


#################################
def readblock(
        f,
        data_type,
        shape=None,
        byteorder=sys.byteorder,
        skip=False,
        htype=np.int32):
    #################################
    """
    data_type = int,float32,float
    or
    data_type = array

    shape	    = tuple
    """

    nbytes_to_read = computeBlockSize(shape, data_type)

    try:
        nb1 = np.frombuffer(f.read(4), htype)
        if sys.byteorder != byteorder:
            nb1.byteswap(True)
        nb1 = nb1[0]

        nbytes = nb1

        # check
        if nbytes_to_read:
            if nbytes_to_read != nbytes:
                print(
                    "inconsistent block header, using nbytes=%d instead" %
                    nbytes_to_read)
                nbytes = nbytes_to_read

    except IndexError:
        raise IOError("ReadBlockError")

    if skip:
        f.seek(nbytes, 1)
        data = None
        shape = None
        print("  skipping %d bytes... " % (nbytes))

    else:

        if isinstance(data_type, tuple):

            data = []
            for tpe in data_type:

                if isinstance(tpe, int):
                    val = f.read(tpe)
                else:
                    bytes = np.dtype(tpe).itemsize
                    val = np.frombuffer(f.read(bytes), tpe)
                    if sys.byteorder != byteorder:
                        val.byteswap(True)

                    val = val[0]

                data.append(val)

        else:
            data = np.frombuffer(f.read(nbytes), data_type)
            if sys.byteorder != byteorder:
                data.byteswap(True)

    nb2 = np.frombuffer(f.read(4), htype)
    if sys.byteorder != byteorder:
        nb2.byteswap(True)
    nb2 = nb2[0]

    if nb1 != nb2:
        raise IOError("ReadBlockError", "nb1=%d nb2=%d" % (nb1, nb2))

    # reshape if needed
    if shape is not None:
        data.shape = shape

    return data


#################################
def ReadBlock(
        f,
        data_type,
        shape=None,
        byteorder=sys.byteorder,
        pio='no',
        htype=np.int32):
    #################################
    """
    data_type = int,float32,float
    or
    data_type = array

    shape	    = tuple

    pio   : parallel io, 'yes' or 'no'
            if 'yes', each proc read each file
            if 'no',  proc 0 read and send to each other

    """

    if mpi.NTask == 1:
        data = readblock(
            f,
            data_type=data_type,
            shape=shape,
            byteorder=byteorder,
            htype=htype)
        return data

    if pio == 'yes':
        data = readblock(
            f,
            data_type=data_type,
            shape=shape,
            byteorder=byteorder,
            htype=htype)
        return data

    else:
        data = mpi.mpi_ReadAndSendBlock(
            f,
            data_type=data_type,
            shape=shape,
            byteorder=byteorder,
            htype=htype)
        return data


#################################
def ReadArray(
        f,
        data_type,
        shape=None,
        byteorder=sys.byteorder,
        pio='no',
        nlocal=None,
        htype=np.int32):
    #################################
    """
    data_type = int,float32,float
    or
    data_type = array

    shape	    = tuple

    """

    if mpi.NTask == 1:
        data = readblock(
            f,
            data_type=data_type,
            shape=shape,
            byteorder=byteorder,
            htype=np.int32)
        return data

    if pio == 'yes':
        data = readblock(
            f,
            data_type=data_type,
            shape=shape,
            byteorder=byteorder,
            htype=np.int32)
        return data

    else:
        data = mpi.mpi_OldReadAndSendArray(
            f,
            data_type,
            shape=shape,
            byteorder=byteorder,
            nlocal=nlocal,
            htype=np.int32)

    return data

#################################


def ReadDataBlock(
        f,
        data_type,
        shape=None,
        byteorder=sys.byteorder,
        pio='no',
        npart=None,
        skip=False):
    #################################
    """

    Read a block containg data.
    If NTask = 1 or  pio = 'yes', the block is read normally.
    If NTask > 1 and pio = 'no',  the master reads the block and send the data to the slaves.


    In the second case :

    a) the master send N/Ntask element to each task.
    b) if the var npart is present, he send Np/Ntask to each task, for each Np of npart.


    data_type = array

    shape	    = tuple

    """

    if mpi.NTask == 1 or pio == 'yes':
        data = readblock(
            f,
            data_type=data_type,
            shape=shape,
            byteorder=byteorder,
            skip=skip)
        return data

    else:
        data = mpi.mpi_ReadAndSendArray(
            f,
            data_type,
            shape=shape,
            byteorder=byteorder,
            npart=npart,
            skip=skip)

    return data


#################################
def writeblock(f, data, byteorder=sys.byteorder, htype=np.int32):
    #################################
    """
    data = array
    or
    data = ((x,float32),(y,int),(z,float32),(label,40))

    shape	    = tuple
    """

    if isinstance(data, tuple):

        # first, compute nbytes
        nbytes = 0
        for dat in data:
            if isinstance(dat[0], bytes) or isinstance(dat[0], str):
                nbytes = nbytes + dat[1]
            else:
                nbytes = nbytes + np.array([
                    dat[0]], dat[1]).dtype().bytes * np.array([
                        dat[0]], dat[1]).size()

        nbytes = np.array([nbytes], htype)

        # write block
        if sys.byteorder != byteorder:
            nbytes.byteswap(True)

        f.write(nbytes.tostring())
        for dat in data:
            if isinstance(dat[0], bytes) or isinstance(dat[0], str):
                f.write(str.ljust(dat[0], dat[1])[:dat[1]])
            else:
                ar = np.array([dat[0]], dat[1])
                if sys.byteorder != byteorder:
                    ar.byteswap(True)
                f.write(ar.tostring())

        f.write(nbytes.tostring())

    else:
        # write block
        #nbytes = array([data.type().bytes*data.size()],int)
        nbytes = np.array([data.nbytes], htype)

        if sys.byteorder != byteorder:
            nbytes.byteswap(True)
            data.byteswap(True)

        f.write(nbytes.tostring())
        f.write(data.tostring())
        f.write(nbytes.tostring())


#################################
def WriteBlock(f, data, byteorder=sys.byteorder, htype=np.int32):
    #################################
    """
    data = ((x,float32),(y,int),(z,float32),(label,40))

    shape	    = tuple
    """

    if f is not None:

        if isinstance(data, tuple):

            # first, compute nbytes
            nbytes = 0
            for dat in data:
                if isinstance(dat[0], str):
                    nbytes = nbytes + dat[1]
                elif isinstance(dat[0], bytes):
                    nbytes = nbytes + dat[1]
                else:
                    nbytes = nbytes + np.array([dat[0]], dat[1]).nbytes
            nbytes = np.array([nbytes], htype)

            # write block
            if sys.byteorder != byteorder:
                nbytes.byteswap(True)

            f.write(nbytes.tostring())
            for dat in data:
                if isinstance(dat[0], bytes):
                    f.write(bytes.ljust(dat[0], dat[1])[:dat[1]])
                elif isinstance(dat[0], str):
                    tmp = dat[0].encode("utf-8")
                    f.write(bytes.ljust(tmp, dat[1])[:dat[1]])
                else:
                    ar = np.array([dat[0]], dat[1])
                    if sys.byteorder != byteorder:
                        ar.byteswap(True)
                    f.write(ar.tostring())

            f.write(nbytes.tostring())


#################################
def WriteArray(
        f,
        data,
        byteorder=sys.byteorder,
        pio='no',
        npart=None,
        htype=np.int32):
    #################################
    """
    data = array

    shape	    = tuple
    """

    if mpi.NTask == 1 or pio == 'yes':
        writeblock(f, data, byteorder=byteorder, htype=htype)

    else:
        mpi.mpi_GatherAndWriteArray(
            f, data, byteorder=byteorder, npart=npart, htype=htype)


#################################
def WriteDataBlock(f, data, byteorder=sys.byteorder, pio='no', npart=None):
    #################################
    """

    Write a block containg data.
    If NTask = 1 or  pio = 'yes', the block is written normally.
    If NTask > 1 and pio = 'no',  the master get the block from the slaves and write it.

    In the second case :

    a) the master get N/Ntask element from each task.
    b) if the var npart is present, he get Np/Ntask from each task, for each Np of npart.


    data = array

    shape	    = tuple
    """

    if mpi.NTask == 1 or pio == 'yes':
        writeblock(f, data, byteorder=byteorder)

    else:
        mpi.mpi_GatherAndWriteArray(f, data, byteorder=byteorder, npart=npart)


###############################################################
#
# some special function reading gadget related files
#
###############################################################


#################################
def read_cooling(file):
    #################################
    """
    Read cooling file
    """

    f = open(file, 'r')
    f.readline()
    f.readline()
    lines = f.readlines()
    f.close()

    lines = list(map(str.strip, lines))
    elts = list(map(str.split, lines))

    logT = np.array([float(x[0]) for x in elts])
    logL0 = np.array([float(x[1]) for x in elts])
    logL1 = np.array([float(x[2]) for x in elts])
    logL2 = np.array([float(x[3]) for x in elts])
    logL3 = np.array([float(x[4]) for x in elts])
    logL4 = np.array([float(x[5]) for x in elts])
    logL5 = np.array([float(x[6]) for x in elts])
    logL6 = np.array([float(x[7]) for x in elts])

    return logT, logL0, logL1, logL2, logL3, logL4, logL5, logL6


#################################
def read_params(file):
    #################################
    """
    Read params Gadget file and return the content in
    a dictionary
    """

    f = open(file, "rb")
    lines = f.readlines()
    f.close()

    # remove empty lines
    lines = [l for l in lines if l != b'\n']

    # remove trailing
    lines = list(map(bytes.strip, lines))

    # remove comments
    lines = [x for x in lines if len(x.decode("latin-1"))>0 and ( x.decode("latin-1")[0] != '%' or x.decode("latin-1")[0] != '#') ]
    
    # split lines
    elts = list(map(bytes.split, lines))
    

    

    # make dictionary
    params = {}
    for e in elts:
            
        name = e[0].decode("latin-1")
        try:
            params[name] = float(e[1])
        except ValueError:
            params[name] = e[1]

    return params


class RockstarReader:
    def __init__(self, filename):
        self.filename = filename

        d = {}

        with open(filename, "r") as f:
            f = [row for row in f if row[0] != '#' or row[:3] == "#ID"]
            reader = csv.DictReader(f, delimiter=' ')
            for i in reader.fieldnames:
                d[i] = []

            for row in reader:
                for i in reader.fieldnames:
                    d[i].append(row[i])
        for key in d:
            d[key] = np.array(d[key], dtype=np.float32)

        self.data = d

    @property
    def pos(self):
        pos = np.array([
            self.data["X"],
            self.data["Y"],
            self.data["Z"]
        ])
        return pos.transpose()

    @property
    def vel(self):
        vel = np.array([
            self.data["VX"],
            self.data["VY"],
            self.data["VZ"]
        ])
        return vel.transpose()

    @property
    def mvir(self):
        return self.data["Mvir"]

    @property
    def rvir(self):
        return self.data["Rvir"]

    @property
    def num(self):
        return self.data["#ID"]

    @property
    def npart(self):
        return self.data["Np"]

    def __getitem__(self, i):
        output = {}
        for key in self.data:
            output[key] = self.data[key][i]
        return output



###############################################################
#
# some special function/class to read LG pickle files
#
###############################################################


class sdict(dict):
  """
  a slightly extended dictionnary class
  """
  
  def selectc(self,c): 
    """
    return a subset of the dictionnary 
    with only elements where the condition is true 
    """ 
    d = copy.deepcopy(self)
    for key in d.keys(): 
      d[key] = np.compress(c,d[key])
    return d  
    
  def insert(self,entry):
    """
    insert a new entry
    the entry must be a dictionary.
    """
    for key in self.keys():
      if key in entry:
        self[key] = np.concatenate((self[key],np.array([entry[key]])))
      else:
        self[key] = np.concatenate((self[key],np.array([None])))
  
    
    

def readLGData(file):
  
  f = open(file,"rb")
  data = pickle.load(f,encoding='latin1')
  f.close()
      
  odata = sdict()  

  names = np.array([])
  for name in data.keys():
    names = np.append(names,name)
    
  odata["name"] = names
    
  attrs =  data[name].keys()

  # loop over 
  for attr_name in attrs:
  
    # create a variable with the name attr
    vect = np.array([],float)
   
    # loop over all galaxies
    for name in names:
       vect = np.append( vect, data[name][attr_name] )
           
    odata[attr_name] = vect
       
  return odata     
     

###############################################################
#
# function to read yaml files
#
###############################################################

def ReadYaml(file_name,parameter_name=None,encoding="latin-1"):
  """
  read a yaml file and return a dictionary
  """

  with open(file_name, mode="rt", encoding=encoding) as file_name:
    data = yaml.safe_load(file_name)

  if parameter_name is not None:
    if parameter_name in data:
      return data[parameter_name]
    else:
      raise KeyError      

  else:
    return data


###############################################################
#
# Single Stellar Population Tables
#
###############################################################


class SSPGrid():

  def __init__(self, filename):
    self.filename = filename
  
  def write(self,Ages,MH,magnitudes=None,luminosities=None,bolometric_luminosities=None,minimfin=None,SunAbsMag=None,cmdline=""):
    
    import h5py
    with h5py.File(self.filename, 'w') as f:
      
      # Header
      f.create_group("Header")
      dset = f["Header"]
      
      # some attributes
      dset.attrs["version"]  = pNbody.__version__
      dset.attrs["author"]   = libutil.get_UserName()
      dset.attrs["date"]     = libutil.get_Date()
      dset.attrs["cmdline"]  = libutil.get_CommandLine()
      dset.attrs['git_tag']  = libutil.get_GitTag()
      
      # sun absolute magnitude in the given filter
      if SunAbsMag is not None:
        dset.attrs['SunAbsMag']     = SunAbsMag
      
      
      # Data
      f.create_group("Data")
      dset = f["Data"]    
      
      data_set = dset.create_dataset('Ages', data=Ages)
      data_set.attrs['units'] = "[Myr]"

      data_set = dset.create_dataset('MH', data=MH)
      data_set.attrs['units'] = "log10([M/H)"
      
      if magnitudes is not None:
        data_set = dset.create_dataset('Magnitudes', data=magnitudes)
        data_set.attrs['units'] = "[magnitudes/IMF_initial_mass]"

      if luminosities is not None:
        data_set = dset.create_dataset('Luminosities', data=luminosities)
        data_set.attrs['units'] = "[log10(Lsol)/IMF_initial_mass]"
        
      if bolometric_luminosities is not None:
        data_set = dset.create_dataset('Bolometric Luminosities', data=bolometric_luminosities)
        data_set.attrs['units'] = "[log10(Lsol)/IMF_initial_mass]"        
        
      if minimfin is not None:
        data_set = dset.create_dataset('MiniMfin', data=minimfin)
        data_set.attrs['units'] = "[-]"
        
                
        
      
  def info(self):
    
    import h5py
    with h5py.File(self.filename, 'r') as f:
      
      content = self.read()
      
      # Header
      hattrs = content["Header"]
      print("version     : %s"%str(hattrs["version"]))
      print("author      : %s"%str(hattrs["author"]))
      print("date        : %s"%str(hattrs["date"]))
      print("cmdline     : %s"%str(hattrs["cmdline"]))
      
      # Data
      dset = content["Data"]
      print()
      print("        Ages: %s"%str(dset["Ages"].shape))
      print("            : %s"%str(dset["Ages_units"]))
      print("         max: %s"%str(dset["Ages"].max()))
      print("         min: %s"%str(dset["Ages"].min())) 
      
      print()
      print("          MH: %s"%str(dset["MH"].shape))
      print("            : %s"%str(dset["MH_units"]))
      print("         max: %s"%str(dset["MH"].max()))
      print("         min: %s"%str(dset["MH"].min())) 
      
      if "Magnitudes" in dset:
        print()
        print("  Magnitudes: %s"%str(dset["Magnitudes"].shape))
        print("         max: %s"%str(dset["Magnitudes"].max()))
        print("         min: %s"%str(dset["Magnitudes"].min()))  
        print("        mean: %s"%str(dset["Magnitudes"].mean()))   
        print("         std: %s"%str(dset["Magnitudes"].std())) 
        print("            : %s"%str(dset["Magnitudes_units"]))


      if "Bolometric Luminosities" in dset:
        print()
        print("Bolo Luminosities: %s"%str(dset["Bolometric Luminosities"].shape))
        print("              max: %s"%str(dset["Bolometric Luminosities"].max()))
        print("              min: %s"%str(dset["Bolometric Luminosities"].min()))
        print("             mean: %s"%str(dset["Bolometric Luminosities"].mean()))   
        print("              std: %s"%str(dset["Bolometric Luminosities"].std()))          
        print("                 : %s"%str(dset["Bolometric Luminosities_units"]))

      if "Luminosities" in dset:
        print()
        print("Luminosities: %s"%str(dset["Luminosities"].shape))
        print("         max: %s"%str(dset["Luminosities"].max()))
        print("         min: %s"%str(dset["Luminosities"].min()))
        print("        mean: %s"%str(dset["Luminosities"].mean()))   
        print("         std: %s"%str(dset["Luminosities"].std()))          
        print("            : %s"%str(dset["Luminosities_units"]))

        
      if "MiniMfin" in dset:
        print()
        print("   MiniMfin : %s"%str(dset["MiniMfin"].shape))
        print("         max: %s"%str(dset["MiniMfin"].max()))
        print("         min: %s"%str(dset["MiniMfin"].min()))
        print("        mean: %s"%str(dset["MiniMfin"].mean()))   
        print("         std: %s"%str(dset["MiniMfin"].std()))     
        print("            : %s"%str(dset["MiniMfin_units"]))
                      
      
      
  def read(self):
    
    import h5py
    with h5py.File(self.filename, 'r') as f:
      
      content = {}
      
      hattrs = f["Header"].attrs
      content["Header"] = {}
      
      content["Header"]["version"] = hattrs["version"]
      content["Header"]["author"]  = hattrs["author"]
      content["Header"]["date"]    = hattrs["date"]
      content["Header"]["cmdline"] = hattrs["cmdline"]
      
      if "SunAbsMag" in hattrs:
        content["Header"]["SunAbsMag"] = hattrs["SunAbsMag"]
      

      # data
      dset = f["Data"]
      content["Data"] = {}
        
      content["Data"]["Ages"]       = dset["Ages"][:]
      content["Data"]["Ages_units"] = dset["Ages"].attrs["units"]
      
      content["Data"]["MH"]         = dset["MH"][:]  
      content["Data"]["MH_units"]   = dset["MH"].attrs["units"]
      
      if "Magnitudes" in  dset: 
        content["Data"]["Magnitudes"] = dset["Magnitudes"][:]
        content["Data"]["Magnitudes_units"] = dset["Magnitudes"].attrs["units"]

      if "Luminosities" in  dset: 
        content["Data"]["Luminosities"] = dset["Luminosities"][:]
        content["Data"]["Luminosities_units"] = dset["Luminosities"].attrs["units"]

      if "Bolometric Luminosities" in  dset: 
        content["Data"]["Bolometric Luminosities"] = dset["Bolometric Luminosities"][:]
        content["Data"]["Bolometric Luminosities_units"] = dset["Bolometric Luminosities"].attrs["units"]


      if "MiniMfin" in  dset: 
        content["Data"]["MiniMfin"] = dset["MiniMfin"][:]
        content["Data"]["MiniMfin_units"] = dset["MiniMfin"].attrs["units"]
        
                
    
    return content  
    
    
        
  
  
  
  
  
  
    










###########################################################################################
#  package:   Mockimgs
#  file:      lib.py
#  brief:     some useful routines
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
from astropy import units as u

from .. import message

def FluxToSurfaceBrightness(FluxMap,object_distance,ccd_pixel_area):
  """
  Convert a mass/luminosity/flux to a surface brightness map
  
  FluxMap         : the flux map (numpy.array)
  object_distance : distance to the object (astropy.units.quantity)
  ccd_pixel_area  : ccd pixel area (astropy.units.quantity)
  
  """
    
  # compute the absolute magnitude in each pixel (as before we ignore the zero point)
  c = FluxMap==0
  FluxMap = np.where(FluxMap==0,1e-40, FluxMap)        
  MagMap = np.where(c,0,-2.5*np.log10(FluxMap))
  
  
  # compute the apparent magnitude in each pixel
  # Mpc -> 10pc
  d = object_distance.to(10*u.pc).value
  magMap = MagMap + 5*np.log10(d)
  
  # compute the surface brightness in each pixel 
  SBMap = np.where(MagMap==0, 100, magMap + 2.5*np.log10(ccd_pixel_area.to(u.arcsec**2).value) )
  
  return SBMap
       


def FilterFluxMap(FluxMap,filter_opts={'mode':None},verbose=0):
  """
  Filter the flux map
  
  FluxMas      : the flux map (numpy.array)
  filter_opts  : filter options
  """

  
  if filter_opts['mode'] == "psf-convolution":
    """
    if a fits file is given as a psf, convolve with it
    """
    from astropy.io import fits
    from scipy import ndimage
    from astropy.convolution import convolve_fft
        
    psf_filename      = filter_opts["psf_filename"]
    image_pixel_scale = filter_opts["image_pixel_scale"]
    
    message.message("convolving with",verbose=verbose,verbosity=1,level=0,color='b')
    message.message("image_pixel_scale = %g"%image_pixel_scale,verbose=verbose,verbosity=1,level=0,color='b')    
    
    # open the psf fits file
    psf      = fits.open(psf_filename)
    psf_data = psf[0].data
    
    psf_header = psf[0].header
    
    if "PIXELSCL" in psf_header:
      psf_pixel_scale=psf_header["PIXELSCL"]
    else:
      message.message("PIXELSCL keywork is missing in %s"%opt.psf_filename,verbose=1,verbosity=0,level=0,color='r')
      message.message("STOP !",verbose=1,verbosity=0,level=0,color='r')
      exit()    
    
    # check if pixel scale agree
    scale_factor = psf_pixel_scale/image_pixel_scale
    message.message("scale_factor = %g"%scale_factor,verbose=verbose,verbosity=1,level=0,color='b')
          
    # new size of the psf
    psf_new_shape = (int(scale_factor*psf_data.shape[0]),int(scale_factor*psf_data.shape[1]))
    
    # rescale the pdf
    psf_data = ndimage.zoom(psf_data, scale_factor, output=None, order=3, mode='constant', cval=-1, prefilter=True)
    
    # do the convolution
    FluxMap = convolve_fft(FluxMap, psf_data, boundary='fill',fill_value=0.0, allow_huge=True, normalize_kernel=True)


  #if filter_opts is not None:
    # here, we should do something
    
    #if opt.convolve=="gaussian":   
    #  print("Convolving with a gaussian...")
    #  psf = opt.psf/opt.ccd_pixel_size[0]
    #  psf_map = gaussian_filter(L_map.shape[0],psf)
    #elif opt.convolve=="square":   
    #  print("Convolving with a square...")
    #  psf_map = square_filter(int(opt.psf))
    #        
    ## convolve
    #L_map = signal.convolve2d(L_map, psf_map, mode='same', boundary='fill', fillvalue=0)
  #  pass
  
  return FluxMap




#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      libmiyamoto.py
#  brief:     Defines Miyamoto profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np



def Potential(M, a, R, z, G=1.):
    """
    Kuzmin Potential
    """
    return -G * M / np.sqrt(R**2 + (a + np.sqrt(z**2))**2)


def Vcirc(M, a, R, G=1.):
    """
    Kuzmin Circular velocity
    """
    return np.sqrt(G * M / (R**2 + a**2)**1.5 * R**2)


def Omega(M, a, R, G=1.):
    """
    Kuzmin-Nagai Omega
    """
    return np.sqrt(G * M / R**3)


def Kappa(M, a, R, G=1.):
    """
    Kuzmin Kappa
    """

    r2 = R * R
    x = np.sqrt(r2 + a**2)
    return np.sqrt(G * M * (-3. * r2 / x**5 + 4. / x**3))


def SurfaceDensity(M, a, R, G=1.):
    """
    Kuzmin Density
    """

    R2 = R * R

    return a*M/(2*np.pi) / (R2 + a**2)**(3/2.)




def dR_Potential(M, a, R, z, G=1.):
    """
    first derivative in R
    """

    R2 = R * R
    zb = np.sqrt(z**2)
    azb2 = (a + zb)**2

    R2azb2 = R2 + azb2

    return G * M * R / R2azb2**(3.0 / 2.0)


def d2R_Potential(M, a, R, z, G=1.):
    """
    second derivative in R
    """

    R2 = R * R
    zb = np.sqrt(z**2)
    azb2 = (a + zb)**2

    R2azb2 = R2 + azb2

    return -3 * G * M * R2 * \
        R2azb2**(-5.0 / 2.0) + G * M * R2azb2**(-3.0 / 2.0)


def dz_Potential(M, a, R, z, G=1.):
    """
    first derivative in R
    """

    R2 = R * R
    zb = np.sqrt(z**2)
    azb2 = (a + zb)**2

    R2azb2 = R2 + azb2

    return G * M * z * (a + zb) / (R2azb2**(3.0 / 2.0) * zb)


def d2z_Potential(M, a, R, z, G=1.):
    """
    second derivative in R
    """

    R2 = R * R
    z2 = z * z
    zb2 = z**2
    zb = np.sqrt(zb2)
    azb = a + zb
    azb2 = azb**2

    R2azb2 = R2 + azb2

    c1 = -3 * z2 * azb2 / (R2azb2**(5.0 / 2.0) * zb2)
    c2 = z2 / (R2azb2**(3.0 / 2.0) * zb2)
    c3 = -  z2 * azb / (R2azb2**(3.0 / 2.0) * zb**3.0)
    c4 = azb / (R2azb2**(3.0 / 2.0) * zb)

    return G * M * (c1 + c2 + c3 + c4)





